from qiskit import QuantumCircuit, transpile
from math import pi

def construct_circuit(operations, k):
    """
    Constructs a quantum circuit based on a list of gate operations and the specified number of qubits.
    """

    # Initialize the quantum circuit with k qubits
    qc = QuantumCircuit(k)

    # Parse and add each operation to the circuit
    for op in operations:
        gate_type, qubits = op

        # Add the corresponding gate to the circuit
        if gate_type == 'CNOT':
            qc.cx(qubits[0]-1, qubits[1]-1)
        elif gate_type == 'CZ':
            qc.cz(qubits[0]-1, qubits[1]-1)
        elif gate_type == 'SWAP':
            qc.swap(qubits[0]-1, qubits[1]-1)
        elif gate_type == 'H':
            qc.h(qubits-1)
        elif gate_type == 'S':
            qc.s(qubits-1)
        elif gate_type == 'Xsqrt':
            qc.sx(qubits-1)
        elif gate_type == 'C(X,X)':
            qc.rxx(pi/2,qubits[0]-1, qubits[1]-1)
        elif gate_type == 'GammaXYZ':
            qc.sdg(qubits-1)
            qc.h(qubits-1)
        elif gate_type == 'GammaXZY':
            qc.h(qubits-1)
            qc.s(qubits-1)
        elif gate_type == 'Z':
            qc.z(qubits-1)
        elif gate_type == 'X':
            qc.x(qubits-1)
        elif gate_type == 'Y':
            qc.y(qubits-1)
        elif gate_type == 'T':
            qc.t(qubits-1)
        elif gate_type == 'CCZ':
            qc.ccz(qubits[0]-1, qubits[1]-1, qubits[2]-1)
        else:
            print(f"Unsupported gate type: {gate_type}")

    return qc


from autqec.utils.linalg import *
from autqec.utils.perms import *
from autqec.utils.pauli import *
from autqec.utils.qec import compute_standard_form
from autqec.utils.symplectic import *
import subprocess
import re
import pickle
from itertools import combinations

class qec_code_ZX_dualities_from_magma_with_intersection:
    def __init__(self,n,k,d,H_symp):
        """
        H = (X | Z)

        Args:
            n (int): number of physical qubits
            k (int): number of logical qubits
            H_symp (np.array): Stabilizers (in binary symplectic format) H = (SX|SZ)
        """

        self.n = n
        self.k = k
        self.d = d
        self.H_symp = H_symp

    def qubits_2bitrep_order(self):
        n = self.n
        qubit_doublets_list = np.arange(1,2*n+1,1) # Qubit triplets
        magma_order = qubit_doublets_list.reshape((n,2)).flatten(order='F') # X | Z 
        X_part = magma_order[:n]
        Z_part = magma_order[n:2*n]
        two_bit_rep_order = np.concatenate((X_part,Z_part)) #  X | Z
        return two_bit_rep_order
    
    def preprocess_H(self):
        n = self.n
        H_rref, _, transform_rows, transform_cols = rref_mod2(self.H_symp)
        qubit_labels_og = self.qubits_2bitrep_order()
        reordered_qubit_list = qubit_labels_og@transform_cols
        return reordered_qubit_list, H_rref, transform_rows, transform_cols

    def convert_to_magma_mat(self,mat,mat_name='M'):
        """
        Args: 
            mat (np.array): matrix 
        """
        mat = np.array(mat,dtype=int)
        n_rows, n_cols = mat.shape
        mat_str = [','.join(map(str, row)) for row in mat]
        mat_str = ',\n'.join(mat_str)
        magma_code = f"{mat_name} := Matrix(GF(2), {n_rows}, {n_cols},\n" + "[" + mat_str + "]);\n"
        return magma_code 

    def run(self,fileroot,save_auts = True,save_magma_commands = False, save_magma_output = False):
        n = self.n
        k = self.k
        d = self.d

        reordered_qubit_list, H_rref, transform_rows, transform_cols = self.preprocess_H()

        H_rref_MAGMA = self.convert_to_magma_mat(H_rref)
        id_mat = np.eye(n,dtype=int) 
        id_row = np.hstack([id_mat,id_mat])@transform_cols
        I_MAGMA = self.convert_to_magma_mat(id_row,mat_name='I')

        commands_part1 = """
        // Define a function to find prime factors
        PrimeFactors := function(n)
            // Use Factorization function to find prime factors
            F := Factorization(n);
            
            // Initialize an empty list to store prime factors
            prime_factors := [];
            
            // Iterate through the factors and extract the primes
            for pair in F do
                Append(~prime_factors, pair[1]);
            end for;
            
            // Return the list of prime factors
            return prime_factors;
        end function;
        """

        commands_part2 = """
        C1 := LinearCode(M);
        C1;
        C2 := LinearCode(I);
        printf "End\n";
        printf "\n";
        time
        autgroup1 := AutomorphismGroup(C1); 
        autgroup2 := AutomorphismGroup(C2);
        autgroup := autgroup1 meet autgroup2;
        autgroup_order := Order(autgroup);
        printf "\n";
        printf "Order: ";
        autgroup_order;
        printf "\n";
        prime_factors := PrimeFactors(autgroup_order);

        for p in prime_factors do
            printf "Sylow Order: ";
            p;
            for g in Generators(Sylow(autgroup,p)) do
                printf "---\n";
                g;
                printf "---\n";
            end for;
        end for;
        """

        commands = commands_part1 + H_rref_MAGMA + I_MAGMA + commands_part2

        if save_magma_commands == True:
            with open(fileroot + f'magma_commands_n{n}k{k}d{d}.txt', "w") as file:
                file.write(commands)
        
        raw_magma_output = self.magma(commands)

        if save_magma_output == True:
            with open(fileroot + f'magma_output_n{n}k{k}d{d}.txt', "w") as file:
                file.write(raw_magma_output)

        # time
        time_pattern = r"Time:\s*([\d\.]+)"
        match = re.search(time_pattern, raw_magma_output)
        if match:
            time = float(match.group(1))
        else:
            time = 0.0
        
        # automorphism group order
        order_pattern = r"Order:\s*(\d+)"
        match = re.search(order_pattern, raw_magma_output)
        if match:
            order = int(match.group(1))
        else:
            order = 1

        # automorphism group generators and qubit relabelling to original basis
        aut_gens, aut_gens_text = self.parse_magma_output_for_aut_gens(raw_magma_output)
        fixed_auts_gens = []
        for g in aut_gens:
            correct_g = []
            for cycle in g:
                new_cycle = []
                for i in cycle:
                    new_cycle.append(reordered_qubit_list[i-1])
                correct_g.append(tuple(new_cycle))
            fixed_auts_gens.append(correct_g)

        # store in dictionary
        code_auts_dict = {}
        code_auts_dict['order'] = order
        code_auts_dict['auts'] = fixed_auts_gens
        code_auts_dict['time'] = time

        if save_auts == True:
            with open(fileroot + f'ZX_dualities_n{n}k{k}d{d}.pkl', 'wb') as file:
                pickle.dump(code_auts_dict, file)

        return code_auts_dict

    def magma(self,commands):
        process = subprocess.Popen(['magma'], stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True)
        stdout, stderr = process.communicate(commands)
        return stdout

    def parse_magma_output_for_aut_gens(self,raw_magma_output):
        """ Extract automorphism group generators """
        aut_gens_text = re.sub(r'\n','',raw_magma_output)
        aut_gens_text = re.sub(r'\s+', ' ', aut_gens_text).strip()
        aut_gens_text = re.findall(r'---(.*?)---',aut_gens_text,re.DOTALL)
        aut_gens = []
        for g in aut_gens_text:
            cycles = g.split(')(')
            one_aut_gen = []
            for cycle in cycles:
                elements = cycle.strip('()').split(',')
                elements = tuple(int(elem) for elem in elements)
                one_aut_gen.append(elements)
            aut_gens.append(one_aut_gen)
        return aut_gens, aut_gens_text
    
class circ_from_ZX_duality:
    def __init__(self,H_symp,aut):
        """
        Class for finding the physical qubit circuits of the 
        generators of the automorphism groups of stabilizer QECCs
        including appropriate Pauli corrections. 

        2-bit representation
        --------------------
        Paulis (X | Z): 
        I --> 00   
        X --> 10   
        Z --> 01   
        
        1-qubit Clifford operators: 
        | Operators      |   Permutation   |
        |----------------|-----------------|
        | $H$            | (1,2)           |

        Args:
            H_symp (np.array): stabilizer generators of the QEC.
            aut (list): list of cycles representing the automorphism.
        """
        if not isinstance(aut, list):
            raise TypeError("Aut must be a list of tuples.")
        if is_matrix_full_rank(H_symp) == False:
            raise AssertionError("Rows of H_symp should be independent. Use a generating set of stabilizers.")
        
        n = H_symp.shape[1]//2
        self.n = n
        self.H_symp = H_symp
        self.aut = aut
        self.bits = np.arange(1,2*n+1,1)
        self.qubit_indices = np.arange(1,n+1,1)
        self.bits_image = permute_by_cycles(self.bits,self.aut)

        # 2-bit rep embedding
        id_mat = np.eye(2*n,dtype=int)
        self.E_mat = id_mat
        self.EInv_mat = id_mat

    def swaps(self): 
        """
        Returns SWAP gates of the automorphism and orders qubits.
        """
        bits_image = self.bits_image

        permuted_qubit_indices = reduce_doublets_to_qubits(bits_image)

        SWAPs_reversed = sort_with_swaps(permuted_qubit_indices)
        physical_SWAP_gates = SWAPs_reversed[::-1]
        
        # Reverse SWAPS.
        qubit_doublets = [(bits_image[i],bits_image[i+1]) for i in range(0, 2*self.n, 2)]
        ordered_qubit_doublets = apply_swaps(qubit_doublets, SWAPs_reversed)
        self.ordered_qubit_triplets = ordered_qubit_doublets

        return physical_SWAP_gates, ordered_qubit_doublets
    
    def single_qubit_cliffords(self,ordered_qubit_doublets):
        """
        Returns the 1-qubit Clifford gates of the automorphism.
        """
        single_qubit_gates = []
        for i,doublet in enumerate(ordered_qubit_doublets):
            gate_ind = i+1
            reduced_triplet = (doublet[0]%2,doublet[1]%2)
            if reduced_triplet == (1,0):
                pass
            elif reduced_triplet == (0,1):
                single_qubit_gates.append(("H",gate_ind))
            else:
                raise AssertionError(f"Unknown triplet: {reduced_triplet}")

        return single_qubit_gates
    
    def auts_to_perm_mat(self):
        """
        Converts a permutation in cyclic notation to a permutation matrix.

        Args:
        perms (list of tuples): List of tuples representing cycles in the permutation.

        Returns:
        np.ndarray: Permutation matrix.
        """     
        # correct qubit order for 2-bit representation (X | Z)
        n = self.n   
        X_bits = [i for i in range(1, 2*n + 1, 2)]
        Z_bits = [i for i in range(2, 2*n + 1, 2)]
        q3bit_order = X_bits + Z_bits
        new_aut = []
        for cycle in self.aut:
            new_aut.append(tuple(q3bit_order.index(x)+1 for x in cycle))
        
        # Initialize the identity matrix of size n
        perm_matrix = np.eye(2*self.n,dtype=int)
        for cycle in new_aut:
            # Rotate the elements in the cycle
            for i in range(len(cycle)):
                from_idx = cycle[i] - 1  # convert to 0-based index
                to_idx = cycle[(i + 1) % len(cycle)] - 1  # next element in the cycle
                perm_matrix[from_idx, from_idx] = 0
                perm_matrix[from_idx, to_idx] = 1
        
        return perm_matrix
    
    def perm_mat_to_symp_mat(self):
        """Take a list of permutations and convert to symplectic matrix by conjugating by E. """
        return np.mod(self.E_mat @ self.auts_to_perm_mat() @ self.EInv_mat,2) 
    
    def symp_transform(self):
        """Check whether matrix A is a direct sum of a matrix M and a symplectic matrix S"""
        symp_mat = self.perm_mat_to_symp_mat()
        if is_symplectic(symp_mat) == False:
            raise AssertionError('Automorphism generator is not a valid symplectic transformation.')
        return symp_mat

    def circ(self):
        """
        Returns the circuit of the automorphism as 
        1-qubit Cliffords + SWAPs.
        """        

        physical_SWAP_gates, ordered_qubit_doublets = self.swaps()
        single_qubit_gates = self.single_qubit_cliffords(ordered_qubit_doublets)
        self.single_qubit_gates = single_qubit_gates
        physical_circuit = single_qubit_gates + physical_SWAP_gates

        return physical_circuit, self.symp_transform()
    




from autqec.automorphisms import *
import subprocess
import re
import pickle

class qec_code_auts_from_magma:
    def __init__(self,n,k,d,H_symp):
        """
        H = (X+Z | X | Z)

        Args:
            n (int): number of physical qubits
            k (int): number of logical qubits
            H_symp (np.array): Stabilizers (in binary symplectic format) H = (SX|SZ)
        """

        self.n = n
        self.k = k
        self.d = d
        self.H_symp = H_symp

        # 3-bit rep of H
        m = H_symp.shape[0]
        id_mat = np.eye(n,dtype=int)
        row_1 = np.hstack((np.hstack((id_mat,id_mat)),id_mat))
        XZ_part = np.zeros((m,n),dtype=int)
        for row_ind in range(m):
            for col_ind in range(n):
                XZ_part[row_ind,col_ind] = (H_symp[row_ind,col_ind] + H_symp[row_ind,col_ind+n])%2
        row_2 = np.hstack((XZ_part,H_symp))
        H_3bit = np.vstack((row_1,row_2))
        self.H_3bit = np.array(H_3bit,dtype=int)
    
    def qubits_3bitrep_order(self):
        n = self.n
        qubit_triplet_list = np.arange(1,3*n+1,1) # Qubit triplets
        magma_order = qubit_triplet_list.reshape((n,3)).flatten(order='F') # Z | X | X+Z
        Z_part = magma_order[:n]
        X_part = magma_order[n:2*n]
        XZ_part = magma_order[2*n:]
        three_bit_rep_order = np.concatenate((XZ_part,X_part,Z_part)) # X+Z | X | Z
        return three_bit_rep_order

    def preprocess_H_3bit(self):
        H_rref, _, transform_rows, transform_cols = rref_mod2(self.H_3bit)
        qubit_labels_og = self.qubits_3bitrep_order()
        reordered_qubit_list = qubit_labels_og@transform_cols
        return reordered_qubit_list, H_rref, transform_rows, transform_cols

    def convert_to_magma_mat(self,mat):
        """
        Args: 
            mat (np.array): matrix 
        """
        mat = np.array(mat,dtype=int)
        n_rows, n_cols = mat.shape
        mat_str = [','.join(map(str, row)) for row in mat]
        mat_str = ',\n'.join(mat_str)
        magma_code = "M := Matrix(GF(2), {}, {},\n".format(n_rows,n_cols) + "[" + mat_str + "]);\n"
        return magma_code 

    def run(self,fileroot,save_auts = True,save_magma_commands = False, save_magma_output = False):
        n = self.n
        k = self.k
        d = self.d

        reordered_qubit_list, H_rref, transform_rows, transform_cols = self.preprocess_H_3bit()

        H_rref_MAGMA = self.convert_to_magma_mat(H_rref)

        commands_part1 = """
        // Define a function to find prime factors
        PrimeFactors := function(n)
            // Use Factorization function to find prime factors
            F := Factorization(n);
            
            // Initialize an empty list to store prime factors
            prime_factors := [];
            
            // Iterate through the factors and extract the primes
            for pair in F do
                Append(~prime_factors, pair[1]);
            end for;
            
            // Return the list of prime factors
            return prime_factors;
        end function;
        """

        commands_part2 = """
        C := LinearCode(M);
        C;
        printf "End\n";
        printf "\n";
        time
        autgroup := AutomorphismGroup(C); 
        autgroup_order := Order(autgroup);
        printf "\n";
        printf "Order: ";
        autgroup_order;
        printf "\n";
        prime_factors := PrimeFactors(autgroup_order);

        for p in prime_factors do
            printf "Sylow Order: ";
            p;
            for g in Generators(Sylow(autgroup,p)) do
                printf "---\n";
                g;
                printf "---\n";
            end for;
        end for;
        """

        commands = commands_part1 + H_rref_MAGMA + commands_part2

        if save_magma_commands == True:
            with open(fileroot + f'magma_commands_n{n}k{k}d{d}.txt', "w") as file:
                file.write(commands)
            print("Run the commands on MAGMA online calculator.")
            print(f"Copy and save output as 'magma_output_n{n}k{k}d{d}.txt'")
            input("Press enter to continue...")
            print("")
            with open(fileroot + f'magma_output_n{n}k{k}d{d}.txt', "r") as file:
                raw_magma_output = file.read()
        else: 
            raw_magma_output = self.magma(commands) # runs MAGMA

        if save_magma_output == True:
            with open(fileroot + f'magma_output_n{n}k{k}d{d}.txt', "w") as file:
                file.write(raw_magma_output)

        # time
        time_pattern = r"Time:\s*([\d\.]+)"
        match = re.search(time_pattern, raw_magma_output)
        if match:
            time = float(match.group(1))
        else:
            time = 0.0
        
        # automorphism group order
        order_pattern = r"Order:\s*(\d+)"
        match = re.search(order_pattern, raw_magma_output)
        if match:
            order = int(match.group(1))
        else:
            order = 1

        # automorphism group generators and qubit relabelling to original basis
        aut_gens, aut_gens_text = self.parse_magma_output_for_aut_gens(raw_magma_output)
        fixed_auts_gens = []
        for g in aut_gens:
            correct_g = []
            for cycle in g:
                new_cycle = []
                for i in cycle:
                    new_cycle.append(reordered_qubit_list[i-1])
                correct_g.append(tuple(new_cycle))
            fixed_auts_gens.append(correct_g)

        # store in dictionary
        code_auts_dict = {}
        code_auts_dict['order'] = order
        code_auts_dict['auts'] = fixed_auts_gens
        code_auts_dict['time'] = time

        if save_auts == True:
            with open(fileroot + f'auts_n{n}k{k}d{d}.pkl', 'wb') as file:
                pickle.dump(code_auts_dict, file)

        return code_auts_dict

    def magma(self,commands):
        process = subprocess.Popen(['magma'], stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True)
        stdout, stderr = process.communicate(commands)
        return stdout

    def parse_magma_output_for_aut_gens(self,raw_magma_output):
        """ Extract automorphism group generators """
        aut_gens_text = re.sub(r'\n','',raw_magma_output)
        aut_gens_text = re.sub(r'\s+', ' ', aut_gens_text).strip()
        aut_gens_text = re.findall(r'---(.*?)---',aut_gens_text,re.DOTALL)
        aut_gens = []
        for g in aut_gens_text:
            cycles = g.split(')(')
            one_aut_gen = []
            for cycle in cycles:
                elements = cycle.strip('()').split(',')
                elements = tuple(int(elem) for elem in elements)
                one_aut_gen.append(elements)
            aut_gens.append(one_aut_gen)
        return aut_gens, aut_gens_text

class qec_code_auts_from_magma_with_intersection:
    def __init__(self,n,k,d,H_symp):
        """
        H = (X+Z | X | Z)

        Args:
            n (int): number of physical qubits
            k (int): number of logical qubits
            H_symp (np.array): Stabilizers (in binary symplectic format) H = (SX|SZ)
        """

        self.n = n
        self.k = k
        self.d = d
        self.H_symp = H_symp

        # 3-bit rep of H
        m = H_symp.shape[0]
        id_mat = np.eye(n,dtype=int)
        row_1 = np.hstack((np.hstack((id_mat,id_mat)),id_mat))
        XZ_part = np.zeros((m,n),dtype=int)
        for row_ind in range(m):
            for col_ind in range(n):
                XZ_part[row_ind,col_ind] = (H_symp[row_ind,col_ind] + H_symp[row_ind,col_ind+n])%2
        row_2 = np.hstack((XZ_part,H_symp))
        H_3bit = np.vstack((row_1,row_2))
        self.H_3bit = np.array(H_3bit,dtype=int)
    
    def qubits_3bitrep_order(self):
        n = self.n
        qubit_triplet_list = np.arange(1,3*n+1,1) # Qubit triplets
        magma_order = qubit_triplet_list.reshape((n,3)).flatten(order='F') # Z | X | X+Z
        Z_part = magma_order[:n]
        X_part = magma_order[n:2*n]
        XZ_part = magma_order[2*n:]
        three_bit_rep_order = np.concatenate((XZ_part,X_part,Z_part)) # X+Z | X | Z
        return three_bit_rep_order

    def preprocess_H_3bit(self):
        H_rref, _, transform_rows, transform_cols = rref_mod2(self.H_3bit)
        qubit_labels_og = self.qubits_3bitrep_order()
        reordered_qubit_list = qubit_labels_og@transform_cols
        return reordered_qubit_list, H_rref, transform_rows, transform_cols

    def convert_to_magma_mat(self,mat,mat_name='M'):
        """
        Args: 
            mat (np.array): matrix 
        """
        mat = np.array(mat,dtype=int)
        n_rows, n_cols = mat.shape
        mat_str = [','.join(map(str, row)) for row in mat]
        mat_str = ',\n'.join(mat_str)
        magma_code = f"{mat_name} := Matrix(GF(2), {n_rows}, {n_cols},\n" + "[" + mat_str + "]);\n"
        return magma_code 

    def run(self,fileroot,save_auts = True,save_magma_commands = False, save_magma_output = False):
        n = self.n
        k = self.k
        d = self.d

        reordered_qubit_list, H_rref, transform_rows, transform_cols = self.preprocess_H_3bit()

        H_rref_MAGMA = self.convert_to_magma_mat(H_rref)
        id_mat = np.eye(n,dtype=int) 
        id_row = np.hstack([id_mat,id_mat,id_mat])@transform_cols
        I_MAGMA = self.convert_to_magma_mat(id_row,mat_name='I')

        commands_part1 = """
        // Define a function to find prime factors
        PrimeFactors := function(n)
            // Use Factorization function to find prime factors
            F := Factorization(n);
            
            // Initialize an empty list to store prime factors
            prime_factors := [];
            
            // Iterate through the factors and extract the primes
            for pair in F do
                Append(~prime_factors, pair[1]);
            end for;
            
            // Return the list of prime factors
            return prime_factors;
        end function;
        """

        commands_part2 = """
        C1 := LinearCode(M);
        C1;
        C2 := LinearCode(I);
        printf "End\n";
        printf "\n";
        time
        autgroup1 := AutomorphismGroup(C1); 
        autgroup2 := AutomorphismGroup(C2);
        autgroup := autgroup1 meet autgroup2;
        autgroup_order := Order(autgroup);
        printf "\n";
        printf "Order: ";
        autgroup_order;
        printf "\n";
        prime_factors := PrimeFactors(autgroup_order);

        for p in prime_factors do
            printf "Sylow Order: ";
            p;
            for g in Generators(Sylow(autgroup,p)) do
                printf "---\n";
                g;
                printf "---\n";
            end for;
        end for;
        """

        commands = commands_part1 + H_rref_MAGMA + I_MAGMA + commands_part2

        if save_magma_commands == True:
            with open(fileroot + f'magma_commands_n{n}k{k}d{d}.txt', "w") as file:
                file.write(commands)
            print("Run the commands on MAGMA online calculator.")
            print(f"Copy and save output as 'magma_output_n{n}k{k}d{d}.txt'")
            input("Press enter to continue...")
            print("")
            with open(fileroot + f'magma_output_n{n}k{k}d{d}.txt', "r") as file:
                raw_magma_output = file.read()
        else: 
            raw_magma_output = self.magma(commands) # runs MAGMA

        if save_magma_output == True:
            with open(fileroot + f'magma_output_n{n}k{k}d{d}.txt', "w") as file:
                file.write(raw_magma_output)

        # time
        time_pattern = r"Time:\s*([\d\.]+)"
        match = re.search(time_pattern, raw_magma_output)
        if match:
            time = float(match.group(1))
        else:
            time = 0.0
        
        # automorphism group order
        order_pattern = r"Order:\s*(\d+)"
        match = re.search(order_pattern, raw_magma_output)
        if match:
            order = int(match.group(1))
        else:
            order = 1

        # automorphism group generators and qubit relabelling to original basis
        aut_gens, aut_gens_text = self.parse_magma_output_for_aut_gens(raw_magma_output)
        fixed_auts_gens = []
        for g in aut_gens:
            correct_g = []
            for cycle in g:
                new_cycle = []
                for i in cycle:
                    new_cycle.append(reordered_qubit_list[i-1])
                correct_g.append(tuple(new_cycle))
            fixed_auts_gens.append(correct_g)

        # store in dictionary
        code_auts_dict = {}
        code_auts_dict['order'] = order
        code_auts_dict['auts'] = fixed_auts_gens
        code_auts_dict['time'] = time

        if save_auts == True:
            with open(fileroot + f'auts_n{n}k{k}d{d}.pkl', 'wb') as file:
                pickle.dump(code_auts_dict, file)

        return code_auts_dict

    def magma(self,commands):
        process = subprocess.Popen(['magma'], stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True)
        stdout, stderr = process.communicate(commands)
        return stdout

    def parse_magma_output_for_aut_gens(self,raw_magma_output):
        """ Extract automorphism group generators """
        aut_gens_text = re.sub(r'\n','',raw_magma_output)
        aut_gens_text = re.sub(r'\s+', ' ', aut_gens_text).strip()
        aut_gens_text = re.findall(r'---(.*?)---',aut_gens_text,re.DOTALL)
        aut_gens = []
        for g in aut_gens_text:
            cycles = g.split(')(')
            one_aut_gen = []
            for cycle in cycles:
                elements = cycle.strip('()').split(',') 
                elements = tuple(int(elem) for elem in elements if 'Id' not in elem)
                one_aut_gen.append(elements)
            aut_gens.append(one_aut_gen)
        return aut_gens, aut_gens_text
    
    


from autqec.automorphisms import *
from autqec.magma_interface import *

n = 5
k = 1 
d = '?'
H_symp = np.load(f'codetables/parity_checks/H_symp_n{n}k{k}.npy')

# without intersection
code_auts_dict = qec_code_auts_from_magma(n,k,d,H_symp).run(fileroot='./',save_auts=False)
auts = code_auts_dict['auts']
G, LX, LZ, D = compute_standard_form(H_symp)
for aut in auts:
    phys_act = circ_from_aut(H_symp,aut)        
    phys_circ,_ = phys_act.circ()
    print(phys_circ)
    logical_act = logical_circ_and_pauli_correct(H_symp,phys_circ)   
    logical_circ = logical_act.run()
    print(logical_circ[0])
print('\n----------------------------\n')
# with intersection
code_auts_dict = qec_code_auts_from_magma_with_intersection(n,k,d,H_symp).run(fileroot='./',save_auts=False)
auts = code_auts_dict['auts']
G, LX, LZ, D = compute_standard_form(H_symp)
for aut in auts:
    phys_act = circ_from_aut(H_symp,aut)        
    phys_circ,_ = phys_act.circ()
    print(phys_circ)
    logical_act = logical_circ_and_pauli_correct(H_symp,phys_circ)   
    logical_circ = logical_act.run()
    print(logical_circ[0])


from autqec.utils.qec import *

# Steane 
def steane():
    print("Steane")
    k = 1
    H_symp = np.vstack((np.array((
                        [0,0,0,1,1,1,1,0,0,0,0,0,0,0],
                        [0,1,1,0,0,1,1,0,0,0,0,0,0,0],
                        [1,0,1,0,1,0,1,0,0,0,0,0,0,0])),
                        np.array((
                        [0,0,0,0,0,0,0,0,0,0,1,1,1,1],
                        [0,0,0,0,0,0,0,0,1,1,0,0,1,1],
                        [0,0,0,0,0,0,0,1,0,1,0,1,0,1]))))
    G, Z_logicals, X_logicals, D = compute_standard_form(H_symp)
    
    # anticommutation of logicals
    assert np.allclose(symp_prod(X_logicals,Z_logicals),np.eye(k))
    # anticommutation of stabilizers & destabilizers
    assert np.allclose(symp_prod(G,D),np.eye(G.shape[0]))
    # commutation of stabilizers with logicals
    for Z in Z_logicals:
        assert np.allclose(symp_prod(G,Z),np.zeros((G.shape[0])))
    for X in X_logicals:
        assert np.allclose(symp_prod(G,X),np.zeros((G.shape[0])))
    # commutation of destabilizers with logicals
    for Z in Z_logicals:
        assert np.allclose(symp_prod(D,Z),np.zeros((D.shape[0])))
    for X in X_logicals:
        assert np.allclose(symp_prod(D,X),np.zeros((D.shape[0])))

# [[5,1,3]]
def n5():
    k=1
    print("5-qubit")
    H_symp = np.array(([[1,0,1,0,1,0,0,1,1,0],
                        [0,0,1,1,0,1,0,0,1,1],
                        [0,1,1,1,1,0,0,0,0,0],
                        [0,0,0,0,0,0,1,1,1,1]]))

    G, Z_logicals, X_logicals, D = compute_standard_form(H_symp)
    
    # anticommutation of logicals
    assert np.allclose(symp_prod(X_logicals,Z_logicals),np.eye(k))
    # anticommutation of stabilizers & destabilizers
    assert np.allclose(symp_prod(G,D),np.eye(G.shape[0]))
    # commutation of stabilizers with logicals
    for Z in Z_logicals:
        assert np.allclose(symp_prod(G,Z),np.zeros((G.shape[0])))
    for X in X_logicals:
        assert np.allclose(symp_prod(G,X),np.zeros((G.shape[0])))
    # commutation of destabilizers with logicals
    for Z in Z_logicals:
        assert np.allclose(symp_prod(D,Z),np.zeros((D.shape[0])))
    for X in X_logicals:
        assert np.allclose(symp_prod(D,X),np.zeros((D.shape[0])))


    # assert np.allclose(symp_prod(S,D),np.eye(S.shape[0]))

# [[15,1,3]]
## REED-MULLER CODE
def reed_muller():
    k=1
    print("Reed-Muller")
    H_X = np.array(([[1,0,0,0,1,1,1,0,0,0,1,1,1,0,1],
                    [0,1,0,0,1,0,0,1,1,0,1,1,0,1,1],
                    [0,0,1,0,0,1,0,1,0,1,1,0,1,1,1],
                    [0,0,0,1,0,0,1,0,1,1,0,1,1,1,1]]))
    H_Z = np.array(([[1,0,0,0,0,0,0,0,0,0,1,1,1,0,0],
                    [0,1,0,0,0,0,0,0,0,0,1,1,0,1,0],
                    [0,0,1,0,0,0,0,0,0,0,1,0,1,1,0],
                    [0,0,0,1,0,0,0,0,0,0,0,1,1,1,0],
                    [0,0,0,0,1,0,0,0,0,0,1,1,0,0,1],
                    [0,0,0,0,0,1,0,0,0,0,1,0,1,0,1],
                    [0,0,0,0,0,0,1,0,0,0,0,1,1,0,1],
                    [0,0,0,0,0,0,0,1,0,0,1,0,0,1,1],
                    [0,0,0,0,0,0,0,0,1,0,0,1,0,1,1],
                    [0,0,0,0,0,0,0,0,0,1,0,0,1,1,1]]))

    H_symp = np.vstack((np.hstack((H_X,np.zeros_like(H_X))),
                np.hstack((np.zeros_like(H_Z),H_Z))))

    G, Z_logicals, X_logicals, D = compute_standard_form(H_symp)
    
    # anticommutation of logicals
    assert np.allclose(symp_prod(X_logicals,Z_logicals),np.eye(k))
    # anticommutation of stabilizers & destabilizers
    assert np.allclose(symp_prod(G,D),np.eye(G.shape[0]))
    # commutation of stabilizers with logicals
    for Z in Z_logicals:
        assert np.allclose(symp_prod(G,Z),np.zeros((G.shape[0])))
    for X in X_logicals:
        assert np.allclose(symp_prod(G,X),np.zeros((G.shape[0])))
    # commutation of destabilizers with logicals
    for Z in Z_logicals:
        assert np.allclose(symp_prod(D,Z),np.zeros((D.shape[0])))
    for X in X_logicals:
        assert np.allclose(symp_prod(D,X),np.zeros((D.shape[0])))

    # print(symp_prod(H_symp,D))


def n5_another():
    k = 1
    print('n5_another')
    H_symp = np.array([
        [1, 0, 1, 0, 0, 0, 0, 0, 1, 1],
        [0, 0, 0, 1, 1, 1, 0, 1, 1, 1],
        [0, 1, 0, 0, 1, 0, 0, 1, 1, 0],
        [0, 0, 1, 1, 0, 0, 1, 1, 1, 1]
    ])
    G, Z_logicals, X_logicals, D = compute_standard_form(H_symp)
    
    # anticommutation of logicals
    assert np.allclose(symp_prod(X_logicals,Z_logicals),np.eye(k))
    # anticommutation of stabilizers & destabilizers
    assert np.allclose(symp_prod(G,D),np.eye(G.shape[0]))
    # commutation of stabilizers with logicals
    for Z in Z_logicals:
        assert np.allclose(symp_prod(G,Z),np.zeros((G.shape[0])))
    for X in X_logicals:
        assert np.allclose(symp_prod(G,X),np.zeros((G.shape[0])))
    # commutation of destabilizers with logicals
    for Z in Z_logicals:
        assert np.allclose(symp_prod(D,Z),np.zeros((D.shape[0])))
    for X in X_logicals:
        assert np.allclose(symp_prod(D,X),np.zeros((D.shape[0])))


steane()
reed_muller()
n5()
n5_another()


from autqec.utils.linalg import *
from autqec.utils.symplectic import *

assert is_identity_matrix(np.eye(5))
assert is_matrix_full_rank(np.array([[1,0],[0,1]])) == True
assert is_matrix_full_rank(np.array([[1,0],[0,1],[0,0]])) == False

M = np.array([[0,0,0,1,1,1,1],
              [0,1,1,0,0,1,1],
              [1,0,1,0,1,0,1]])
M_rref, r, R, C = rref_mod2(M)
print(M_rref)
assert np.allclose(M_rref,R@M@C%2)



from autqec.utils.linalg import *
from autqec.utils.perms import *
from autqec.utils.pauli import *
from autqec.utils.qec import compute_standard_form
from autqec.utils.symplectic import *

class circ_from_aut:
    def __init__(self,H,aut):
        """
        Class for finding the physical qubit circuits of the 
        generators of the automorphism groups of stabilizer QECCs. 

        3-bit representation
        --------------------
        Paulis (Z | X | X+Z): 
        I --> 000   
        X --> 011   
        Z --> 101   
        Y --> 110   
        
        1-qubit Clifford operators: 
        | Operators      |   Permutation   |
        |----------------|-----------------|
        | $H$            | (1,2)           |
        | $S$            | (1,3)           |
        | $\sqrt{X}$     | (2,3)           |
        | $\Gamma_{XZY}$ | (1,2,3)         |
        | $\Gamma_{XYZ}$ | (1,3,2)         |

        Args:
            H (np.array): stabilizer generators of the QEC.
            aut (list): list of cycles representing the automorphism.
        """
        if is_matrix_full_rank(H) == False:
            raise AssertionError("Rows of H should be independent. Use a generating set of stabilizers.")
        if not isinstance(aut, list):
            raise TypeError("Aut must be a list of tuples.")
        n = H.shape[1]//2
        self.n = n
        self.H = H
        self.aut = aut
        self.bits = np.arange(1,3*n+1,1)
        self.qubit_indices = np.arange(1,n+1,1)
        self.bits_image = permute_by_cycles(self.bits,self.aut)

        # 3-bit rep embedding
        E = np.array([[1,1,1],[1,0,1],[1,1,0]],dtype=int)
        EInv = np.mod(np.linalg.inv(E),2)
        EInv = np.array(EInv,dtype=int)
        id_mat = np.eye(n,dtype=int)
        self.E_mat = np.kron(E,id_mat)   
        self.EInv_mat = np.kron(EInv,id_mat)

    def swaps(self): 
        """
        Returns SWAP gates of the automorphism and orders qubits.
        """
        bits_image = self.bits_image
        permuted_qubit_indices = reduce_triplets_to_qubits(bits_image)

        SWAPs_reversed = sort_with_swaps(permuted_qubit_indices)
        physical_SWAP_gates = SWAPs_reversed[::-1]
        
        # Reverse SWAPS.
        qubit_triplets = [tuple(bits_image[i:i+3]) for i in range(0, len(bits_image), 3)]
        ordered_qubit_triplets = apply_swaps(qubit_triplets, SWAPs_reversed)
        self.ordered_qubit_triplets = ordered_qubit_triplets

        return physical_SWAP_gates, ordered_qubit_triplets
    
    def single_qubit_cliffords(self,ordered_qubit_triplets):
        """
        Returns the 1-qubit Clifford gates of the automorphism.
        """
        single_qubit_gates = []
        for i,triplet in enumerate(ordered_qubit_triplets):
            gate_ind = i+1
            reduced_triplet = (triplet[0]%3,triplet[1]%3,triplet[2]%3)
            if reduced_triplet == (1,2,0):
                pass
            elif reduced_triplet == (1,0,2):
                single_qubit_gates.append(("Xsqrt",gate_ind)) 
            elif reduced_triplet == (2,1,0):
                single_qubit_gates.append(("H",gate_ind))
            elif reduced_triplet == (0,2,1):
                single_qubit_gates.append(("S",gate_ind))
            elif reduced_triplet == (0,1,2):
                single_qubit_gates.append(("GammaXZY",gate_ind)) # REVERSE
            elif reduced_triplet == (2,0,1):
                single_qubit_gates.append(("GammaXYZ",gate_ind)) # REVERSE
            else:
                raise AssertionError(f"Unknown triplet: {reduced_triplet}")

        return single_qubit_gates
    
    def auts_to_perm_mat(self):
        """
        Converts a permutation in cyclic notation to a permutation matrix.

        Args:
        perms (list of tuples): List of tuples representing cycles in the permutation.

        Returns:
        perm_matrix (np.array): Permutation matrix.
        """     
        # correct qubit order for 3-bit representation (X+Z | Z | X)
        n = self.n   
        XZ_bits = [i for i in range(3, 3*n + 1, 3)]
        Z_bits = [i for i in range(1, 3*n + 1, 3)]
        X_bits = [i for i in range(2, 3*n + 1, 3)]
        q3bit_order = XZ_bits + Z_bits + X_bits
        new_aut = []
        for cycle in self.aut:
            new_aut.append(tuple(q3bit_order.index(x)+1 for x in cycle))
        
        # Initialize the identity matrix of size n
        perm_matrix = np.eye(3*self.n,dtype=int)
        for cycle in new_aut:
            # Rotate the elements in the cycle
            for i in range(len(cycle)):
                from_idx = cycle[i] - 1  # convert to 0-based index
                to_idx = cycle[(i + 1) % len(cycle)] - 1  # next element in the cycle
                perm_matrix[from_idx, from_idx] = 0
                perm_matrix[from_idx, to_idx] = 1
        
        return perm_matrix
    
    def perm_mat_to_symp_mat(self):
        """Take a list of permutations and convert to symplectic matrix by conjugating by E. """
        return np.mod(self.E_mat @ self.auts_to_perm_mat() @ self.EInv_mat,2) 
    
    def symp_transform(self):
        """Check whether matrix A is a direct sum of a matrix M and a symplectic matrix S"""
        A = self.perm_mat_to_symp_mat()
        n = len(A) // 3
        symp_mat = A[n:,n:]
        if np.sum(A[n:,:n])!=0:
            raise AssertionError('Off-diagonal check: bottom left.')
        elif np.sum(A[:n,n:])!=0:
            raise AssertionError('Off-diagonal check: top right.')
        elif is_symplectic(symp_mat) == False:
            raise AssertionError('Automorphism generator is not a valid symplectic transformation.')
        return symp_mat

    def circ(self):
        """
        Returns the circuit of the automorphism as 
        1-qubit Cliffords + SWAPs.
        """        

        physical_SWAP_gates, ordered_qubit_triplets = self.swaps()
        single_qubit_gates = self.single_qubit_cliffords(ordered_qubit_triplets)
        self.single_qubit_gates = single_qubit_gates
        physical_circuit = single_qubit_gates + physical_SWAP_gates

        return physical_circuit, self.symp_transform()


class logical_circ_and_pauli_correct:
    def __init__(self, H, phys_circ):
        """
        Class for finding appropriate Pauli corrections of 
        Clifford circuits to preserve stabilizers and infer
        the corresponding logical action of the circuit. 

        Args: 
            H (np.array): stabilizer generators of the QEC.
            phys_circ (list): quantum circuit.
        """

        # standard form 
        G, LX, LZ, D = compute_standard_form(H)
        self.G = G
        self.LX = LX
        self.D = D
        self.LZ = LZ

        # code parameters
        n = G.shape[1] // 2
        m = G.shape[0]
        self.n = n 
        self.m = m
        self.k = n-m

        # tableux
        tableux = np.vstack([G,LX,D,LZ]) 
        T_symp_prod, omega = symp_prod(tableux,tableux,return_omega=True)
        self.omega = omega
        if np.allclose(T_symp_prod,omega) == False:
            raise AssertionError("Check stabilizer/destabilizer tableux.")
        self.tableux = tableux
        self.tableux_phases ,self.tableux_pauli = binary_vecs_to_paulis(tableux,phase_bit=True)
        self.show_tableux = '\n'.join([' '.join(x) for x in binary_vecs_to_paulis(tableux)])

        # physical circuit
        self.phys_circ = phys_circ
        
    def new_tableux(self):
        T_G_L = np.vstack([self.G,self.LX,self.LZ])
        T_G_L_4bit = op_2bit_to_op_3bit_and_phase(T_G_L)
        return clifford_circ_stab_update(T_G_L_4bit,self.phys_circ)
    
    def new_tableux_anticomm(self):
        m = self.m
        n = self.n
        k = self.k
        omega = self.omega
        b = np.mod(self.new_tableux()[1][:,n:] @ omega @ self.tableux.T @ omega,2)
        G_comp = b[:m].copy()
        L_comp = b[m:].copy()

        if np.allclose(b[:,n:-k],np.zeros((m+2*k,m),dtype=int)) == False:
            raise AssertionError('Physical circuit maps operators outside of code space.')

        if np.allclose(G_comp[:,m:],np.zeros((m,2*n-m),dtype=int)) == False:
            raise AssertionError('Physical circuit does not preserve the stabilizer group.')
        
        return b, G_comp, L_comp
    
    def U_logical_act(self):
        m = self.m
        k = self.k
        _, G_comp, L_comp = self.new_tableux_anticomm()
        u_act = np.zeros((2*k,2*k),dtype=int)
        u_act[:,:k] = L_comp[:,m:m+k]
        u_act[:,k:] = L_comp[:,-k:]

        return u_act
    
    def i_phases(self,U_ACT):
        """
            Returns a phase vector that operators pick up via U_ACT.
            Phases: {0,1,2,3} == {+1, +i, -1, -i}
        """
        k = len(U_ACT)//2
        p = np.zeros(2*k,dtype=int)
        for row in range(2*k):
            for i in range(k):
                if U_ACT[row,i] == 1 and U_ACT[row,i+k] == 1:
                    p[row] = (p[row] + 1)%4
        return p
    
    def new_tableux_pauli_prod_phases(self):
        n = self.n
        m = self.m
        k = self.k
        tableux = self.tableux
        tableux_pauli = self.tableux_pauli
        tableux_phases = self.tableux_phases

        b,_,_ = self.new_tableux_anticomm()
        tableux_new_phases = np.zeros(len(b),dtype=int)
        for t_ind, t in enumerate(b):
            t_indices = np.where(t==1)[0]
            pauli_multiply = ['I']*self.n
            pauli_phase = 0
            for i in t_indices:
                pauli_multiply, p = multiply_pauli_strings(pauli_multiply,pauli_phase,tableux_pauli[i],tableux_phases[i])
                tableux_new_phases[t_ind] = np.mod(tableux_new_phases[t_ind] + p,4)
        self.i_phases_from_Y_ops = self.i_phases(self.U_logical_act())
        tableux_new_phases[m:] = (tableux_new_phases[m:].copy()+self.i_phases_from_Y_ops)%4
        return tableux_new_phases
                 
    def run(self):
        """ Finds the required Pauli corrections X or Z to fix any -1 phases of stabilizers. """
        U_p = np.zeros(2*self.n,dtype=int)
        pauli_circ = []
        n = self.n
        m = self.m
        k = self.k
        T = self.tableux

        # find phase differences. 
        p = self.new_tableux()[0]
        q = self.new_tableux_pauli_prod_phases()
        phase_diff = np.mod(p - q, 4)
        if np.all(phase_diff % 2 == 0) == False: # check there are no i or -i phases. 
            raise AssertionError("Pauli correction to the physical circuit failed.")

        # correct stabilizer phases. 
        for i in range(self.m):
            if phase_diff[i] != 0:
                h = (i+n)%(2*n)
                U_p = np.mod(U_p + T[h],2)

        # multiply X/Y/Z to logical circuit to match the phase to logical act. 
        ## X-logicals of the code
        for i in range(m,m+k):
            if phase_diff[i] != 0:
                h = (i+n)%(2*n)
                U_p = np.mod(U_p + T[h],2)
        ## Z-logicals of the code
        for i in range(m+k,m+2*k):
            if phase_diff[i] != 0:
                h = (i+m+n)%(2*n)
                U_p = np.mod(U_p + T[h],2)

        # convert U_p into quantum circuit
        for i in range(n):
            if U_p[i] == 1 and U_p[i+n] == 0:
                pauli_circ.append(('X',i+1))
            elif U_p[i] == 0 and U_p[i+n] == 1:
                pauli_circ.append(('Z',i+1))
            elif U_p[i] == 1 and U_p[i+n] == 1:
                pauli_circ.append(('Y',i+1))

        U_ACT = self.U_logical_act()
        circ_logical_act = circ_from_symp_mat(U_ACT).run()
        if np.allclose(symp_mat_prods(circ_logical_act,self.k),U_ACT) == False:
            raise AssertionError("Logical circuit is wrong.")
        return circ_logical_act, pauli_circ + self.phys_circ
    

class circ_from_symp_mat:
    def __init__(self,symp_mat):
        """
        Class to find the quantum circuit implemented by a symplectic transformation 
        up to a logical Pauli correction. 

        Gates: {H, CNOT, CZ, S, Xsqrt, CX(X,X)}
        """
        assert is_symplectic(symp_mat)

        k = symp_mat.shape[0] // 2

        # invert symplectic matrix
        omega = np.eye(2*k,dtype=int)
        omega[:,:k], omega[:,k:] = omega[:,k:].copy(), omega[:,:k].copy()
        self.k = k
        self.symp_mat = omega@symp_mat.T@omega

    def H_gates(self):
        """ Return H gates. """
        symp_mat = self.symp_mat.copy()
        k = self.k

        XX_part = symp_mat[:k,:k].copy()
        XZ_part = symp_mat[:k,k:].copy()
        XX_part_rref, pivots, row_transform, col_transform = rref_mod2(XX_part)
        rk_XX = len(pivots)
        if rk_XX == k:
            return []
        symp_mat_top_half_new = np.hstack([row_transform@XX_part%2,row_transform@XZ_part%2])
        B2 = symp_mat_top_half_new[rk_XX:,k:].copy()
        _, H_pivots,_,_ = rref_mod2(B2)

        return H_pivots
    
    def decomp(self):
        symp_mat = self.symp_mat.copy()
        k = self.k
        self.H_pivots = self.H_gates()
        diag_h = diag_mat(k,self.H_pivots)
        id_m_diag_h = np.eye(k,dtype=int) - diag_h

        U_H = np.zeros_like(symp_mat)
        U_H[:k,:k] = id_m_diag_h
        U_H[:k,k:] = diag_h
        U_H[k:,:k] = diag_h
        U_H[k:,k:] = id_m_diag_h

        symp_mat_new = symp_mat @ U_H 

        C = symp_mat_new[:k,:k].copy()
        A = symp_mat_new[k:,:k].copy() @ inv_mod2(C) % 2
        B = symp_mat_new[:k,k:].copy() @ C.T % 2

        id_mat = np.eye(2*k,dtype=int)
        U_A = id_mat.copy()
        U_A[k:,:k] = A.copy()
        U_B = id_mat.copy()
        U_B[:k,k:] = B.copy()
        U_C = np.zeros_like(id_mat)
        U_C[:k,:k] = C.copy()
        U_C[k:,k:] = inv_mod2(C.copy()).T
        
        return U_A, U_B, U_C, U_H
    
    def run(self):
        U_A, U_B, U_C, U_H = self.decomp()

        # H gates 
        H_circ = [('H',i+1) for i in self.H_pivots]
    
        # CNOT gates 
        k = self.k
        CNOT_circ, reduced_mat = rref_mod2(U_C[:k,:k],CNOTs=True) # need to invert CNOT gates
        assert is_identity_matrix(reduced_mat)

        # S / CZ gates
        XZ_circ = set()
        B = U_B[:k,k:]
        for i in range(k):
            if B[i, i] == 1:
                XZ_circ.add(('S',i+1))
            for j in range(i+1,k):
                if i != j and B[i, j] == 1:
                    XZ_circ.add(('CZ',(i+1,j+1)))

        # Xsqrt / CX(X,X) gates 
        ZX_circ = set()
        A = U_A[k:,:k]
        for i in range(k):
            if A[i, i] == 1:
                ZX_circ.add(('Xsqrt',i+1))
            for j in range(i+1,k):
                if i != j and A[i, j] == 1:
                    ZX_circ.add(('C(X,X)',(i+1,j+1)))

        logical_circ = []
        logical_circs = [H_circ,CNOT_circ[::-1],XZ_circ,ZX_circ]

        for circ in logical_circs:
            if circ:
                logical_circ.extend(circ) 
        
        return logical_circ


import numpy as np

def symp_prod(A,B,return_omega=False):
    """
    Computes the binary symplectic product between A and B.
    """
    A = np.array(A,dtype=int)
    B = np.array(B,dtype=int)

    if A.ndim == 1:
        A = A.reshape(1, -1)
    if B.ndim == 1:
        B = B.reshape(1, -1)
        
    m, n = A.shape
    m_b, n_b = B.shape

    assert n == n_b 

    omega = np.eye(n,dtype=int)
    nhalf = n//2
    omega[:,:nhalf], omega[:,nhalf:] = omega[:,nhalf:].copy(), omega[:,:nhalf].copy()

    if return_omega:
        return np.array((A @ omega @ B.T)%2,dtype=int), omega
    else: 
        return np.array((A @ omega @ B.T)%2,dtype=int)

def is_symplectic(M):
    """
    Asserts that matrix M is symplectic.
    """
    n = M.shape[0]
    prod, omega = symp_prod(M,M,return_omega=True)
    return np.allclose(prod,omega)

def H_gate(i,n):
    i = i - 1
    eye = np.eye(2*n,dtype=int)
    eye[:,[i,i+n]] = eye[:,[i+n,i]]
    return eye

def CZ_gate(i,j,n):
    i = i - 1
    j = j - 1
    eye = np.eye(2*n,dtype=int)
    eye[:,i+n] = eye[:,j] + eye[:,i+n]
    eye[:,j+n] = eye[:,i] + eye[:,j+n]
    return eye

def S_gate(i,n):
    i = i - 1
    eye = np.eye(2*n,dtype=int)
    eye[:,i+n] = eye[:,i] + eye[:,i+n]
    return eye

def Xsqrt_gate(i,n):
    i = i - 1
    eye = np.eye(2*n,dtype=int)
    eye[:,i] = eye[:,i] + eye[:,i+n]
    return eye

def CX_XX_gate(i,j,n):
    i = i - 1
    j = j - 1
    eye = np.eye(2*n,dtype=int)
    eye[:,i] = eye[:,i] + eye[:,j+n]
    eye[:,j] = eye[:,j] + eye[:,i+n]
    return eye

def SWAP_gate(i,j,n):
    i = i - 1
    j = j - 1
    eye = np.eye(2*n,dtype=int)
    eye[:,[i,j]] = eye[:,[j,i]]
    eye[:,[i+n,j+n]] = eye[:,[j+n,i+n]]
    return eye

def CNOT_gate(i,j,n):
    i = i - 1
    j = j - 1
    eye = np.eye(2*n,dtype=int)
    eye[:,j] = eye[:,i] + eye[:,j]
    eye[:,i+n] = eye[:,i+n] + eye[:,j+n]
    return eye

def gamma_XZY_gate(i,n):
    return H_gate(i,n)@S_gate(i,n)

def gamma_XYZ_gate(i,n):
    return S_gate(i,n)@H_gate(i,n)

def symp_mat_prods(ops,n):
    """
    Returns symplectic product of quantum gates.
    """
    mat = np.eye(2*n,dtype=int)
    for op in ops:
        gate_type, qubits = op
        # Add the corresponding gate to the circuit
        if gate_type == 'CNOT':
            mat = mat @ CNOT_gate(qubits[0], qubits[1], n) 
        elif gate_type == 'CZ':
            mat = mat @ CZ_gate(qubits[0], qubits[1], n) 
        elif gate_type == 'SWAP':
            mat = mat @ SWAP_gate(qubits[0], qubits[1], n) 
        elif gate_type == 'H':
            mat = mat @ H_gate(qubits, n) 
        elif gate_type == 'S':
            mat = mat @ S_gate(qubits, n) 
        elif gate_type == 'Xsqrt':
            mat = mat @ Xsqrt_gate(qubits, n) 
        elif gate_type == 'C(X,X)':
            mat = mat @ CX_XX_gate(qubits[0], qubits[1], n) 
        elif gate_type == 'GammaXYZ':
            mat = mat @ gamma_XYZ_gate(qubits, n)
        elif gate_type == 'GammaXZY':
            mat = mat @ gamma_XZY_gate(qubits, n) 
        else:
            raise TypeError(f"Unsupported gate type: {gate_type}")
    return mat%2

# def op_2bit_to_op_3bit(op_2bit):
#     m = op_2bit.shape[0]
#     n = op_2bit.shape[1]//2
#     zeros = np.zeros((m,n),dtype=int)
#     op_3bit = np.hstack((zeros,op_2bit))
#     for row in range(m):
#         for i in range(n):
#             op_3bit[row,i] = (op_3bit[row,i+n]+op_3bit[row,i+2*n])%2
    
#     return np.array(op_3bit,dtype=int)

def op_2bit_to_op_3bit_and_phase(op_2bit):
    if op_2bit.ndim == 1:
        op_2bit = op_2bit.reshape(1, -1)
    m = op_2bit.shape[0]
    n = op_2bit.shape[1]//2
    zeros = np.zeros((m,n),dtype=int)
    op_3bit = np.hstack((zeros,op_2bit))
    phases = np.zeros(m,dtype=int)
    for row in range(m):
        p = 0
        for i in range(n):
            if op_3bit[row,i+n] == 1 and op_3bit[row,i+2*n] == 1:
                p += 1
            op_3bit[row,i] = (op_3bit[row,i+n]+op_3bit[row,i+2*n])%2
        phases[row] = p
    
    return phases, op_3bit

def clifford_circ_stab_update(op_3bit_p,physical_circ):
    """ 
        Args: 
            op_3bit_p (tuple): phases + op_3bit 
    """
    phases = op_3bit_p[0].copy()
    op_3bit = op_3bit_p[1].copy()
    if op_3bit.ndim == 1:
        op_3bit = op_3bit.reshape(1, -1)
    n = op_3bit.shape[1] // 3
    for gate in physical_circ:
        gate_type, qubits = gate
        if type(qubits) == int:
            q = qubits - 1
            a = q # x+z
            b = q + n # x 
            c = q + 2*n # z
        elif type(qubits) == tuple: 
            q1, q2 = qubits
            q1 = q1-1
            q2 = q2-1
        if gate_type == 'X':
            phases = (phases + 2*op_3bit[:,c])%4
        elif gate_type == 'Z':
            phases = (phases + 2*op_3bit[:,b])%4
        elif gate_type == 'S':
            phases = (phases + op_3bit[:,b])%4
            op_3bit[:,[a,c]] = op_3bit[:,[c,a]]
        elif gate_type == 'H':
            phases = (phases - op_3bit[:,a] + op_3bit[:,b] + op_3bit[:,c])%4
            op_3bit[:,[b,c]] = op_3bit[:,[c,b]]
        elif gate_type == 'GammaXYZ':
            phases = (phases + op_3bit[:,a] - op_3bit[:,c])%4
            col_a = op_3bit[:,a].copy() # x+z
            col_b = op_3bit[:,b].copy() # x
            col_c = op_3bit[:,c].copy() # z
            op_3bit[:,a] = col_c
            op_3bit[:,b] = col_a
            op_3bit[:,c] = col_b
        elif gate_type == 'GammaXZY':
            phases = (phases + op_3bit[:,a] - op_3bit[:,b])%4
            col_a = op_3bit[:,a].copy() # x+z
            col_b = op_3bit[:,b].copy() # x
            col_c = op_3bit[:,c].copy() # z
            op_3bit[:,a] = col_b
            op_3bit[:,b] = col_c
            op_3bit[:,c] = col_a
        elif gate_type == 'Xsqrt':
            phases = (phases - op_3bit[:,c])%4
            op_3bit[:,[a,b]] = op_3bit[:,[b,a]]
        elif gate_type == 'SWAP':
            op_3bit[:,[q1,q2]] = op_3bit[:,[q2,q1]]
            op_3bit[:,[q1+n,q2+n]] = op_3bit[:,[q2+n,q1+n]]
            op_3bit[:,[q1+2*n,q2+2*n]] = op_3bit[:,[q2+2*n,q1+2*n]]
        elif gate_type == 'CZ':
            x1 = op_3bit[:,q1+n]
            x2 = op_3bit[:,q2+n]
            z1 = op_3bit[:,q1+2*n]
            z2 = op_3bit[:,q2+2*n]
            phases = (phases + 2*x1*x2)%4
            op_3bit[:,q1] = (op_3bit[:,q1] + op_3bit[:,q2+n])%2
            op_3bit[:,q2] = (op_3bit[:,q2] + op_3bit[:,q1+n])%2
            op_3bit[:,q1+2*n] = (op_3bit[:,q1+2*n] + op_3bit[:,q2+n])%2
            op_3bit[:,q2+2*n] = (op_3bit[:,q2+2*n] + op_3bit[:,q1+n])%2
        elif gate_type == 'CNOT':
            x1 = op_3bit[:,q1+n]
            x2 = op_3bit[:,q2+n]
            z1 = op_3bit[:,q1+2*n]
            z2 = op_3bit[:,q2+2*n]
            op_3bit[:,q1] =  (op_3bit[:,q1] + op_3bit[:,q2+2*n])%2
            op_3bit[:,q2] = (op_3bit[:,q2] + op_3bit[:,q1+n])%2
            op_3bit[:,q2+n] = (op_3bit[:,q2+n] + op_3bit[:,q1+n])%2
            op_3bit[:,q1+2*n] = (op_3bit[:,q1+2*n] + op_3bit[:,q2+2*n])%2
        elif gate_type == 'C(X,X)':
            x1 = op_3bit[:,q1+n]
            x2 = op_3bit[:,q2+n]
            z1 = op_3bit[:,q1+2*n]
            z2 = op_3bit[:,q2+2*n]
            phases = (phases + 2*z1*z2)%4
            op_3bit[:,q1] = (op_3bit[:,q1] + op_3bit[:,q2+2*n])%2
            op_3bit[:,q2] = (op_3bit[:,q2] + op_3bit[:,q1+2*n])%2
            op_3bit[:,q1+n] = (op_3bit[:,q1+n] + op_3bit[:,q2+2*n])%2
            op_3bit[:,q2+n] = (op_3bit[:,q2+n] + op_3bit[:,q1+2*n])%2
        else: 
            raise TypeError(f'Gate type unknown: {gate_type}')
    return phases, op_3bit


import numpy as np
def binary_vecs_to_paulis(mat,phase_bit = False):
    """Returns the rows of a binary-symplectic matrix as Pauli operators."""
    
    if mat.ndim == 1:
        mat = mat.reshape(1, -1)

    n = mat.shape[1] // 2
    m = mat.shape[0]

    pauli_mat = []
    phase = np.zeros(m,dtype=int)
    for i in range(m):
        row_result = []
        for j in range(n):
            pair = (mat[i, j], mat[i, j+n])
            if pair == (0, 0):
                row_result.append('I')
            elif pair == (1, 0):
                row_result.append('X')
            elif pair == (0, 1):
                row_result.append('Z')
            elif pair == (1, 1):
                if phase_bit == True:
                    row_result.append('XZ')
                    phase[i] += 1
                else: 
                    row_result.append('Y')
        pauli_mat.append(row_result)
        
    if phase_bit == True: 
        return phase, pauli_mat
    else:
        return pauli_mat


def multiply_pauli_strings(pauli1, phase1, pauli2, phase2):
    """ Returns the product of two pauli strings accounting for anticommutation of X and Z.
        e.g. Y = (['X','Z'],1) 

        Phases: {0,1,2,3} == {+1, +i, -1, -i}
    """
    if len(pauli1) != len(pauli2):
        raise ValueError("Pauli strings must be of the same length")
    
    result = []
    phase = phase1%4 + phase2%4 # initial phase difference  
    
    for p1, p2 in zip(pauli1, pauli2):
        if p1 == 'I':
            result.append(p2)
        elif p2 == 'I':
            result.append(p1)
        elif p1 == 'X':
            if p2 == 'X': 
                result.append('I')
            elif p2 == 'Z':
                result.append('XZ')
            elif p2 == 'XZ':
                result.append('Z')
        elif p1 == 'Z':
            if p2 == 'X': 
                result.append('XZ')
                phase += 2
            elif p2 == 'Z':
                result.append('I')
            elif p2 == 'XZ':
                result.append('X')
                phase += 2
        elif p1 == 'XZ':
            if p2 == 'X': 
                result.append('Z')
                phase += 2
            elif p2 == 'Z':
                result.append('X')
            elif p2 == 'XZ':
                result.append('I')
                phase += 2
    
    return result, phase%4



import igraph as ig
import numpy as np
from sympy.combinatorics import Permutation, PermutationGroup
import scipy.sparse as sparse

def sparse_pcm_to_tanner_graph(pcm):
    """
    Creates a Tanner graph in igraph format from a sparse parity check matrix.
    Args:
        pcm: A (sparse) parity check matrix.
    Returns:
        An igraph Graph object representing the Tanner graph.
    """
    rows, cols = pcm.shape
    g = ig.Graph()
    g.add_vertices(cols + rows)
    # g.vs["type"] = ["variable"] * cols + ["check"] * rows
    g.vs["color"] = [1] * cols + [2] * rows
    edges = []
    rows_indices, col_indices = pcm.nonzero()
    for i in range(len(rows_indices)):
        edges.append((col_indices[i], cols + rows_indices[i]))
    g.add_edges(edges)
    return g

def graph_aut_group(g, n, print_order=True):
    """
    Calculates and returns the automorphism group of an igraph graph as a sympy PermutationGroup.

    Args:
        g: An igraph Graph object.
        n: number of columns of adjancency matrix
        print_order: A boolean indicating whether to print the order of the automorphism group.
                     Defaults to True.

    Returns:
        A sympy PermutationGroup object representing the automorphism group of the graph,
        or Identity if no automorphism generators are found.
    """
    automorphism_generators = g.automorphism_group(color=g.vs["color"])
    if automorphism_generators:

        sympy_permutations = [Permutation(list(generator)[:n]) for generator in automorphism_generators]
        sympy_group = PermutationGroup(sympy_permutations)
        if print_order:
            print("Automorphism group order:",sympy_group.order())
        return sympy_group
    else:
        print("Cannot create sympy group, no automorphism generators found.")
        return PermutationGroup(Permutation(range(1)))


def B_for_all_cliffords(n,bits_3=True):
    """ Parity-check matrix for classical code B that has 
        all possible Clifford gates as automorphisms.
    """
    
    identity = np.eye(n,dtype=np.uint8)
    if bits_3:
        return np.hstack([identity,identity,identity])
    else:
        v = []
        for i in range(n):
            v.extend([i,n+i])
        return np.hstack([identity,identity])
    
def valid_clifford_auts(pcm,bits_3=True,return_order=True):
    m,pcm_cols = pcm.shape
    
    if bits_3:
        n = pcm_cols//3
        v = np.column_stack([np.arange(n), np.arange(n, 2*n), np.arange(2*n, 3*n)]).flatten()
        B_pcm = B_for_all_cliffords(n,bits_3)[:,v]
    else: 
        n = pcm_cols//2
        v = np.column_stack([np.arange(n), np.arange(n, 2*n)]).flatten()
        B_pcm = B_for_all_cliffords(n,bits_3)[:,v]

    B_tanner_graph = sparse_pcm_to_tanner_graph(B_pcm)
    B_graphauts = graph_aut_group(B_tanner_graph,pcm_cols,False)

    pcm = pcm[:,v]
    tanner_graph = sparse_pcm_to_tanner_graph(pcm)
    code_graph_auts = graph_aut_group(tanner_graph,pcm_cols,False)
    clifford_auts = B_graphauts.subgroup_search(code_graph_auts.__contains__)
    if return_order: 
        print(f'Order: {clifford_auts.order()}')
    
    clifford_auts_cyclic = []
    for a in clifford_auts:
        clifford_auts_cyclic.append(a.cyclic_form)
   
    return cycle_indices(clifford_auts_cyclic)

def valid_clifford_auts_B_rows(pcm,bits_3=True,return_order=True):
    m,pcm_cols = pcm.shape
    
    if bits_3:
        n = pcm_cols//3
        v = np.column_stack([np.arange(n), np.arange(n, 2*n), np.arange(2*n, 3*n)]).flatten()
        B_pcm = B_for_all_cliffords(n,bits_3)[:,v]
    else: 
        n = pcm_cols//2
        v = np.column_stack([np.arange(n), np.arange(n, 2*n)]).flatten()
        B_pcm = B_for_all_cliffords(n,bits_3)[:,v]

    pcm = np.vstack([pcm[:,v],B_pcm])
    tanner_graph = sparse_pcm_to_tanner_graph(pcm)
    code_graph_auts = graph_aut_group(tanner_graph,pcm_cols,False)

    if return_order: 
        print(f'Order: {code_graph_auts.order()}')
    
    auts_cyclic = []
    for a in code_graph_auts:
        auts_cyclic.append(a.cyclic_form)
   
    return cycle_indices(auts_cyclic)

def cycle_indices(lists):
    return [[[x + 1 for x in sublist] for sublist in group] for group in lists]


import itertools
from math import comb
from autqec.utils.linalg import *

class qec_embed_code:
    def __init__(self, H, embedding = 'two_code_blocks', custom_embedding = None):
        """ Class for embedding stabilizer codes into a binary linear code 
            where extra columns correspond to pairs of qubits.
            
            Action of CNOT, CZ gates can represented as automorphisms. 
            
            User can restrict connectivity by only adding columns for connected 
            pairs of qubits. 

            Args: 
                H (np.array): stabilizer generators of the original QEC.
                embedding (str): options are 'two_code_blocks', 'all_weight_2', 'custom'
                custom_embedding (np.array): binary matrix representing qubit connectivities, e.g. np.array([[1,1,0],[0,1,1]])
            """
        self.H = H 
        n = H.shape[1]//2
        self.n = n
        self.embedding = embedding
        self.custom_embedding = custom_embedding

    def gen_two_block_embed_mat(self):
        nhalf = self.n//2
        id_mat = np.eye(nhalf,dtype=int)
        zeros_mat = np.zeros_like(id_mat,dtype=int)

        block_col1 = np.hstack((id_mat,zeros_mat))
        block_col2 = np.hstack((zeros_mat,id_mat))
        block_col3 = np.hstack((id_mat,id_mat))

        return np.vstack((np.vstack((block_col1,block_col2)),block_col3)).T

    def gen_w2_embed_mat(self):
        n = self.n
        combos = itertools.combinations(range(n), 2)
        w2_embed_mat = np.zeros((comb(n,2),n),dtype=int)
        for row, pair in enumerate(combos):
            w2_embed_mat[row,pair[0]] = 1
            w2_embed_mat[row,pair[1]] = 1
        return np.vstack((np.eye(n,dtype=int),w2_embed_mat)).T

    def embed_mat(self):
        H = self.H
        n = self.n
        option = self.embedding
        
        HX_part = H[:,:n]
        HZ_part = H[:,n:]
        if option == 'two_code_blocks':
            V_T = self.gen_two_block_embed_mat()
        elif option == 'all_weight_2': 
            V_T = self.gen_w2_embed_mat()
        elif option == 'custom':
            V_T = self.custom_embedding.T
        else:
            raise TypeError("Unknown option for embedding.")
        M_T = V_T[:,n:]
        M = M_T.T
        A_V = np.hstack((HX_part,(HX_part@M_T)%2))
        ones = np.eye(M.shape[0])
        zeros = np.zeros((HZ_part.shape[0],ones.shape[1]))
        B_V = np.vstack((np.hstack((HZ_part,zeros)),np.hstack((M,ones))))
        assert A_V.shape[1] == B_V.shape[1]
        no_of_rows = B_V.shape[0] - A_V.shape[0] 
        A_V = np.vstack((A_V,np.zeros((no_of_rows,A_V.shape[1]))))
        return np.array(np.hstack((A_V,B_V)),dtype=int)
    
def embed_circ_to_normal_circ(embed_circ, n_og, embed_codespace_dict):
    new_circuit = []
    for gate in embed_circ:
        gate_type, q_ind = gate
        if gate_type == 'SWAP':
            q1 = q_ind[0]
            q2 = q_ind[1]
            if q1 < n_og + 1 and q2 < n_og + 1:
                new_circuit.append(gate)
            elif q1 > n_og and q2 > n_og:
                embed_col1 = embed_codespace_dict[int(q1 - n_og)]
                embed_col2 = embed_codespace_dict[int(q2 - n_og)]
                q_a, q_b = embed_col1
                q_c, q_d = embed_col2
                new_circuit.extend([('CNOT',(q_b,q_a)),('CNOT',(q_d,q_c)),('SWAP',(q_a,q_c)),
                                    ('CNOT',(q_b,q_a)),('CNOT',(q_d,q_c))]) 
            elif q1 < n_og+1 and q2 > n_og:
                embed_col = embed_codespace_dict[int(q2 - n_og)]
                if q1 in embed_col: 
                    targ = q1
                    ctrl = [x for x in embed_col if x != q1][0]
                    new_circuit.append(('CNOT',(ctrl,targ)))
                else: 
                    q_a,q_b = embed_col
                    new_circuit.extend([('SWAP',(q1,q_a)),('CNOT',(q_b,q_a)),('SWAP',(q1,q_a))])
            elif q1 > n_og and q2 < n_og+1:
                embed_col = embed_codespace_dict[int(q1 - n_og)]
                if q2 in embed_col: 
                    targ = q2
                    ctrl = [x for x in embed_col if x != q2][0]
                    new_circuit.append(('CNOT',(ctrl,targ)))
                else: 
                    q_a,q_b = embed_col
                    new_circuit.extend([('SWAP',(q2,q_a)),('CNOT',(q_b,q_a)),('SWAP',(q2,q_a))])
        else: 
            if q_ind < n_og + 1:
                new_circuit.append(gate)
            else: 
                if gate_type == 'S':
                    pair = embed_codespace_dict[int(q_ind - n_og)]
                    new_circuit.append(('S',pair[0]))
                    new_circuit.append(('S',pair[1]))
                    new_circuit.append(('CZ',(pair[0],pair[1])))
    
    return new_circuit       


import numpy as np
from autqec.utils.linalg import rref_mod2, inv_mod2
from autqec.utils.symplectic import symp_prod
from autqec.utils.linalg import is_matrix_full_rank

def compute_standard_form(G):
    """
    Returns the standard form of a stabilizer code. 
    See Nielsen & Chuang Section 10.5.7.
    """
    if is_matrix_full_rank(G) == False:
        raise AssertionError("Rows of G should be independent. Use a generating set of stabilizers.")
    n, m = G.shape[1] // 2, G.shape[0]
    k = n-m
    
    G1 = G[:, :n]
    G2 = G[:, n:]

    ## LOGICALS

    # Step 1: Gaussian Elimination on G1 & adjust G2 rows & cols accordingly
    G1_rref, r_pivots, G1_transform_rows, G1_transform_cols = rref_mod2(G1)
    r = len(r_pivots)
    G = np.hstack((G1_rref,(G1_transform_rows@G2@G1_transform_cols)%2))

    # Step 2: Swap columns r to n from X to Z block 
    G[:,r:n], G[:,n+r:2*n] = G[:,n+r:2*n].copy(), G[:,r:n].copy()
    E1 = G[:, :n]
    E2 = G[:, n:]
    E1_rref, e_pivots, E_transform_rows, E_transform_cols = rref_mod2(E1)
    e = len(e_pivots)
    s = e - r
    G = np.hstack((E1_rref,(E_transform_rows@E2@E_transform_cols)%2))
    G[:,r:n], G[:,n+r:2*n] = G[:,n+r:2*n].copy(), G[:,r:n].copy()

    # Step 3: Z Logicals
    A_2 = G[:r,n-k:n]
    C = G[:r,(2*n-k):]
    E = G[r:,(2*n-k):]

    r1 = A_2.T
    r2 = np.zeros((k,n-k-r))
    r3 = np.eye(k)
    right = np.hstack((np.hstack((r1,r2)),r3))
    Z_logicals = np.hstack((np.zeros((k,n)),right))
    Z_logicals = np.array(Z_logicals,dtype=int)
    for z in Z_logicals:
        assert len(z) == 2*n
        assert np.allclose(symp_prod(G,z),np.zeros((G.shape[0])))

    # Step 4: X Logicals
    l1 = np.zeros((k,r))
    l2 = E.T
    l3 = np.eye(k)
    left = np.hstack((np.hstack((l1,l2)),l3))
    r1 = C.T
    r2 = np.zeros((k,n-k-r))
    r3 = np.zeros((k,k))
    right = np.hstack((np.hstack((r1,r2)),r3))
    X_logicals = np.hstack((left,right))
    X_logicals = np.array(X_logicals,dtype=int)
    for x in X_logicals:
        assert len(x) == 2*n
        assert np.allclose(symp_prod(G,x),np.zeros((G.shape[0])))

    # Step 5: Move columns (but not rows) back to their original position
    inv_E_transform_cols = inv_mod2(E_transform_cols)
    inv_G1_transform_cols = inv_mod2(G1_transform_cols)
    inv_transform_cols = inv_E_transform_cols @ inv_G1_transform_cols
    ## STABILIZERS
    G_new = np.zeros_like(G)
    G_new[:,:n] = (G[:,:n] @ inv_transform_cols)%2
    G_new[:,n:] = (G[:,n:] @ inv_transform_cols)%2
    ## LOGICALS
    Z_logicals_og_basis = np.zeros_like(Z_logicals)
    X_logicals_og_basis = np.zeros_like(X_logicals)
    Z_logicals_og_basis[:,:n] =  (Z_logicals[:,:n] @ inv_transform_cols)%2
    Z_logicals_og_basis[:,n:] =  (Z_logicals[:,n:] @ inv_transform_cols)%2
    X_logicals_og_basis[:,:n] =  (X_logicals[:,:n] @ inv_transform_cols)%2
    X_logicals_og_basis[:,n:] =  (X_logicals[:,n:] @ inv_transform_cols)%2
    ## DESTABILIZERS 
    DX = np.hstack([np.zeros((r,n)),np.eye(r),np.zeros((r,n-r))])
    DZ = np.hstack([np.zeros((s,r)),np.eye(s),np.zeros((s,n+k))])
    D = np.array(np.vstack([DX,DZ]),dtype=int)
    D_og_basis = np.zeros_like(D)
    D_og_basis[:,:n] =  (D[:,:n] @ inv_transform_cols)%2
    D_og_basis[:,n:] =  (D[:,n:] @ inv_transform_cols)%2

    return G_new, X_logicals_og_basis, Z_logicals_og_basis, D_og_basis

def stabs_to_H_symp(stabs):
    m = len(stabs)
    n = len(stabs[0])
    H_symp = np.zeros((m,2*n),dtype=int)
    for i_row, s in enumerate(stabs): 
        for i_col, pauli in enumerate(s):
            if pauli == 'I':
                pass
            elif pauli == 'X':
                H_symp[i_row,i_col] = 1
            elif pauli == 'Y':
                H_symp[i_row,i_col] = 1
                H_symp[i_row,i_col+n] = 1
            elif pauli == 'Z':
                H_symp[i_row,i_col+n] = 1
            else: 
                raise TypeError('Unknown Pauli: ',pauli)
            
    return H_symp


import numpy as np
from autqec.utils.perms import permute_by_cycles, reduce_triplets_to_qubits
import pytest

# permute_by_cycles()
n = 4
ordered_list = np.arange(1,3*n+1,1)
# test q1 and q2 swap 
perm1 = [(1,4),(2,5),(3,6)]
permuted_list = permute_by_cycles(ordered_list,perm1).tolist()
assert permuted_list == [4,5,6,1,2,3,7,8,9,10,11,12]

#############################################
assert reduce_triplets_to_qubits(permuted_list) == [2, 1, 3, 4]

permuted_list = [1,5,3,4,2,6]
def test_wrong_triplets():
    with pytest.raises(AssertionError):
        reduce_triplets_to_qubits(permuted_list)




from autqec.ZX_dualities import *
from autqec.utils.qec import stabs_to_H_symp

n = 5
k = 1 
d = 3
stabs = ['XZZXI','IXZZX','XIXZZ','ZXIXZ']
H_symp = stabs_to_H_symp(stabs)
G,LX,LZ,D = compute_standard_form(H_symp)
code = qec_code_ZX_dualities_from_magma_with_intersection(n,k,d,H_symp)
auts_data = code.run('./tests/',save_auts=False)
auts = auts_data['auts']
#######################################################################

phys_act = circ_from_ZX_duality(H_symp,auts[1])
bits_image = phys_act.bits_image
circ, _ = phys_act.circ()
print(circ)


#######################################################################
from autqec.automorphisms import *
circ = logical_circ_and_pauli_correct(H_symp,circ).run()
print(circ)


import subprocess
import numpy as np

def convert_to_magma_mat(mat,mat_name='M'):
    """
    Args: 
        mat (np.array): matrix 
    """
    mat = np.array(mat,dtype=int)
    n_rows, n_cols = mat.shape
    mat_str = [','.join(map(str, row)) for row in mat]
    mat_str = ',\n'.join(mat_str)
    magma_code = f"{mat_name} := Matrix(GF(2), {n_rows}, {n_cols},\n" + "[" + mat_str + "]);\n"
    return magma_code 

def magma(commands):
    process = subprocess.Popen(['magma'], stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True)
    stdout, stderr = process.communicate(commands)
    return stdout

def MAGMA_define_SG(symplectic_mat_list,A):
    """Checks if A is in the span of a list of symplectic matrices."""
    k = symplectic_mat_list[0].shape[0]//2
    commands = ''
    listofgens = '['
    for i, sympmat in enumerate(symplectic_mat_list):
        commands += (convert_to_magma_mat(sympmat,'a'+f'{i}'))
        listofgens += f'a{i},'
    listofgens = listofgens[:-1]
    listofgens += ']'
    A_magma = convert_to_magma_mat(A,'A')
    commands += f"""
listofgens := {listofgens};
G := GeneralLinearGroup({2*k},GF(2));
listminimal := [];
indices := [];
s := 1;
order_previous := 1; // Order of the trivial subgroup
while s le #listofgens do
    x := listofgens[s];
    Append(~listminimal, x);
    S := sub<G | listminimal>;
    order_current := #S;
    if order_current gt order_previous then
        Append(~indices, s);
        order_previous := order_current;
    else
        Prune(~listminimal); // Remove the last element from listminimal
    end if;
    s := s + 1;
end while;
SG := sub<G | listminimal>;
counter := 1;
command := "FG<";
for i in indices do
    if counter lt #indices then
        command cat:= "a";
        command cat:= IntegerToString(indices[counter]);
        command cat:=",";
        counter := counter+1;
    else
        command cat:="a";
        command cat:= IntegerToString(i);
        command cat:= ">, phi := FPGroup(SG); return FG, phi";
    end if;
end for;
FG, phi := eval command;
order:=1;
order:=Order(FG);
printf "Order: ";
order;
structure:=GroupName(FG);
printf "Structure: ";
structure;
if order eq 1 then 
    exit;
end if;
{A_magma}
A in SG;
        """
    return commands



from autqec.utils.pauli import *


mat = np.array([[1,0,1,1]]) #YZ
assert binary_vecs_to_paulis(mat,phase_bit = False)[0][0] == 'Y'
assert binary_vecs_to_paulis(mat,phase_bit = False)[0][1] == 'Z'

mat = np.array([[1,0,1,1]]) #YZ
assert binary_vecs_to_paulis(mat,phase_bit = True)[0] == 1
assert binary_vecs_to_paulis(mat,phase_bit = True)[1][0][0] == 'XZ'
assert binary_vecs_to_paulis(mat,phase_bit = True)[1][0][1] == 'Z'

pauli1 = ['XZ','Z']
phase1 = 1
pauli2 = ['X','Z']
phase2 = 0
pauli_product = multiply_pauli_strings(pauli1, phase1, pauli2, phase2)

assert pauli_product[0][0] == 'Z'
assert pauli_product[0][1] == 'I'
assert pauli_product[1] == 3 # phase (power of i)


from autqec.automorphisms import circ_from_symp_mat
from autqec.utils.symplectic import *
from autqec.utils.linalg import inv_mod2

def random_symp_mats(h):
    _,n = h.shape
    X_part = h[:,:n//2].copy()
    Z_part = h[:,n//2:].copy()
    h_ZX = np.hstack((Z_part,X_part))
    symp_mat = (np.eye(n,dtype=int) + h_ZX.T@h)%2
    assert is_symplectic(symp_mat)
    return symp_mat

n = 20
for _ in range(10000):
    h = np.random.randint(0, 2, n).reshape((1,n))
    S1 = random_symp_mats(h)
    h = np.random.randint(0, 2, n).reshape((1,n))
    S2 = random_symp_mats(h)
    h = np.random.randint(0, 2, n).reshape((1,n))
    S3 = random_symp_mats(h)
    h = np.random.randint(0, 2, n).reshape((1,n))
    S4 = random_symp_mats(h)
    h = np.random.randint(0, 2, n).reshape((1,n))
    S5 = random_symp_mats(h)
    h = np.random.randint(0, 2, n).reshape((1,n))
    S6 = random_symp_mats(h)

    S_in = np.mod(S1 @ S2 @ S3 @ S4 @ S5 @ S6,2)

    gates = circ_from_symp_mat(S_in).run()
    S_out = symp_mat_prods(gates,n//2)
    assert np.allclose(S_in,S_out)



import pickle
import re
import subprocess
from autqec.utils.symplectic import *
from itertools import combinations, permutations
import os
from autqec.utils.magma import convert_to_magma_mat

class clifford_subgroups:
    def __init__(self,k,symplectic_mat_list):
        """
        Args: 
            k (int): number of logical qubits
            symplectic_mat_dict (dict): dictionary of subgroup generators given as symplectic matrices.
        """
        self.k = k
        self.symplectic_mat_list = symplectic_mat_list

    def MAGMA_define_SG(self):
        k = self.k
        symplectic_mat_list = self.symplectic_mat_list
        commands = ''
        listofgens = '['
        for i, sympmat in enumerate(symplectic_mat_list):
            commands += (convert_to_magma_mat(sympmat,'a'+f'{i}'))
            listofgens += f'a{i},'
        listofgens = listofgens[:-1]
        listofgens += ']'
        commands += f"""
listofgens := {listofgens};
G := GeneralLinearGroup({2*k},GF(2));
listminimal := [];
indices := [];
s := 1;
order_previous := 1; // Order of the trivial subgroup
while s le #listofgens do
    x := listofgens[s];
    Append(~listminimal, x);
    S := sub<G | listminimal>;
    order_current := #S;
    if order_current gt order_previous then
        Append(~indices, s);
        order_previous := order_current;
    else
        Prune(~listminimal); // Remove the last element from listminimal
    end if;
    s := s + 1;
end while;
SG := sub<G | listminimal>;
counter := 1;
command := "FG<";
for i in indices do
    if counter lt #indices then
        command cat:= "a";
        command cat:= IntegerToString(indices[counter]);
        command cat:=",";
        counter := counter+1;
    else
        command cat:="a";
        command cat:= IntegerToString(i);
        command cat:= ">, phi := FPGroup(SG); return FG, phi";
    end if;
end for;
FG, phi := eval command;
order:=1;
order:=Order(FG);
printf "Order: ";
order;
structure:=GroupName(FG);
printf "Structure: ";
structure;
if order eq 1 then 
    exit;
end if;
        """
        return commands
    
    def return_order(self):
        commands = self.MAGMA_define_SG()
        process = subprocess.Popen(['magma'], stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True)

        stdout, stderr = process.communicate(commands)
        lines = stdout.strip().split('\n')

        for line in lines:
            if line.startswith("Order:"):
                # Extract the order
                order = int(line.split(":")[1].strip())
                return order
        else:
            return 1
    
    def return_structure(self):
        commands = self.MAGMA_define_SG()
        process = subprocess.Popen(['magma'], stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True)

        stdout, stderr = process.communicate(commands)
        lines = stdout.strip().split('\n')

        for line in lines:
            if line.startswith("Structure:"):
                # Extract the order
                structure = line.split(":")[1].strip()
                return structure
        else:
            return 1
    
    def individual_SWAPs(self):
        k = self.k
        SWAP_gates = ""
        gates_var_names = "["
        gates_str = "["
        combos = list(combinations(range(k), 2))
        for pair in combos:
            i = pair[0]
            j = pair[1]
            SWAP_gates += convert_to_magma_mat(SWAP_gate(i+1,j+1,k),f"SWAP_{i+1}_{j+1}")
            gates_var_names += f"SWAP_{i+1}_{j+1},"
            gates_str += f"\"SWAP_({i+1},{j+1})\","
        gates_var_names = gates_var_names[:-1]
        gates_var_names += "]"
        gates_str = gates_str[:-1]
        gates_str += "]"
        return SWAP_gates, gates_var_names, gates_str

    def individual_CNOTs(self):
        k = self.k
        CNOT_gates = ""
        gates_var_names = "["
        gates_str = "["
        combos = list(permutations(range(k), 2))
        for pair in combos:
            i = pair[0]
            j = pair[1]
            CNOT_gates += convert_to_magma_mat(CNOT_gate(i+1,j+1,k),f"CNOT_{i+1}_{j+1}")
            gates_var_names += f"CNOT_{i+1}_{j+1},"
            gates_str += f"\"CNOT_({i+1},{j+1})\","
        gates_var_names = gates_var_names[:-1]
        gates_var_names += "]"
        gates_str = gates_str[:-1]
        gates_str += "]"
        return CNOT_gates, gates_var_names, gates_str

    def individual_CZs(self):
        k = self.k
        CZ_gates = ""
        gates_var_names = "["
        gates_str = "["
        combos = list(combinations(range(k), 2))
        for pair in combos:
            i = pair[0]
            j = pair[1]
            CZ_gates += convert_to_magma_mat(CZ_gate(i+1,j+1,k),f"CZ_{i+1}_{j+1}")
            gates_var_names += f"CZ_{i+1}_{j+1},"
            gates_str += f"\"CZ_({i+1},{j+1})\","
        gates_var_names = gates_var_names[:-1]
        gates_var_names += "]"
        gates_str = gates_str[:-1]
        gates_str += "]"
        return CZ_gates, gates_var_names, gates_str

    def full_single_qubit_gamma_XYZ(self):
        k = self.k
        gamma_gates = ""
        gates_str = "["
        gates_var_names = "["
        for i in range(k):
            gamma_gates += convert_to_magma_mat(gamma_XYZ_gate(i+1,k),f"gammaXYZ{i+1}")
            gates_str += f"\"gammaXYZ_{i+1}\","
            gates_var_names += f"gammaXYZ{i+1},"
        gates_str = gates_str[:-1]
        gates_var_names = gates_var_names[:-1]
        gates_str += "]"
        gates_var_names += "]"
        return gamma_gates, gates_var_names, gates_str
    
    def full_single_qubit_gamma_XZY(self):
        k = self.k
        gamma_gates = ""
        gates_str = "["
        gates_var_names = "["
        for i in range(k):
            gamma_gates += convert_to_magma_mat(gamma_XZY_gate(i+1,k),f"gammaXZY{i+1}")
            gates_str += f"\"gammaXZY_{i+1}\","
            gates_var_names += f"gammaXZY{i+1},"
        gates_str = gates_str[:-1]
        gates_var_names = gates_var_names[:-1]
        gates_str += "]"
        gates_var_names += "]"
        return gamma_gates, gates_var_names, gates_str

    def full_SWAP_network(self):
        k = self.k
        SWAP_network = ""
        gates_var_names = "["
        gates_str = "["
        for i in range(k-1):
            j = i+1
            SWAP_network += convert_to_magma_mat(SWAP_gate(i+1,j+1,k),f"SWAP_{i+1}_{j+1}")
            gates_var_names += f"SWAP_{i+1}_{j+1},"
            gates_str += f"\"SWAP_({i+1},{j+1})\","
        gates_var_names = gates_var_names[:-1]
        gates_var_names += "]"
        gates_str = gates_str[:-1]
        gates_str += "]"
        return SWAP_network, gates_var_names, gates_str

    def full_CNOT_network(self): 
        k = self.k
        if k == 1:
            return "  "
        # check 1 CNOT gate: CNOT_(O,1)
        ctrl = 0
        targ = 1
        one_CNOT_gate = convert_to_magma_mat(CNOT_gate(ctrl+1,targ+1,k),"CNOT_1_2")
        # check full SWAP network
        SWAP_network, SWAP_var_names, SWAP_gates_str = self.full_SWAP_network()

        return one_CNOT_gate + SWAP_network, "[CNOT_1_2," + SWAP_var_names[1:], "[\"CNOT_(1,2)\"," + SWAP_gates_str[1:]
    
    def full_single_qubit_H(self):
        k = self.k
        H_gates = ""
        gates_str = "["
        gates_var_names = "["
        for i in range(k):
            gate = H_gate(i+1,k)
            H_gates += convert_to_magma_mat(gate,f"H{i+1}")
            gates_str += f"\"H_{i+1}\","
            gates_var_names += f"H{i+1},"
        gates_str = gates_str[:-1]
        gates_var_names = gates_var_names[:-1]
        gates_str += "]"
        gates_var_names += "]"
        return H_gates, gates_var_names, gates_str
    
    def full_single_qubit_S(self):
        k = self.k
        S_gates = ""
        gates_str = "["
        gates_var_names = "["
        for i in range(k):
            gate = S_gate(i+1,k)
            S_gates += convert_to_magma_mat(gate,f"S{i+1}")
            gates_str += f"\"S_{i+1}\","
            gates_var_names += f"S{i+1},"
        gates_str = gates_str[:-1]
        gates_var_names = gates_var_names[:-1]
        gates_str += "]"
        gates_var_names += "]"
        return S_gates, gates_var_names, gates_str
        
    def full_clifford(self):
        k = self.k
        H_gates, H_var_names, H_gates_str = self.full_single_qubit_H() 
        S_gates, S_var_names, S_gates_str = self.full_single_qubit_S()
        if k == 1:
            return H_gates + S_gates, H_var_names[:-1] + ',' + S_var_names[1:], H_gates_str[:-1] + ',' + S_gates_str[1:]
        CNOT_network, CNOT_network_var_names, CNOT_network_str = self.full_CNOT_network()
        gates = H_gates + S_gates + CNOT_network
        gates_var_names = H_var_names[:-1] + ',' + S_var_names[1:-1] + ',' + CNOT_network_var_names[1:]
        gates_as_str = H_gates_str[:-1] + ',' + S_gates_str[1:-1] + ',' + CNOT_network_str[1:]
        return gates, gates_var_names, gates_as_str
        
    def MAGMA_check_gates(self,gates, gates_var_names, gates_as_str):
        magma_commands = self.MAGMA_define_SG()
        cliffList = f"cliffList:={gates_var_names};\n"
        cliffListStr = f"cliffListStr:={gates_as_str};"
        check_membership = """
index := 1;
printf "Logical operators: \\n";
for cliff in cliffList do
    if cliff in SG then
        print cliffListStr[index], cliff @@ phi;
    end if;
index := index+1;
end for;
printf "FINISH";
"""
        commands = magma_commands + gates + cliffList + cliffListStr + check_membership 
        process = subprocess.Popen(['magma'], stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True)

        stdout, stderr = process.communicate(commands)
        lines = stdout.strip().split('\n')

        logical_operators = {}
        logical_ops_flag = False
        for line in lines:
            if line.startswith("Order:"):
                # Extract the order
                order = int(line.split(":")[1].strip())
            elif line.startswith("Logical operators:"):
                # Skip the line that states "Logical operators:"
                logical_ops_flag = True
                continue
            if logical_ops_flag == True:
                # Extract logical operators
                parts = line.split(maxsplit=1)
                if len(parts) == 2:
                    key, value = parts
                    logical_operators[key] = value
            if line.startswith("FINISH"):
                break
        
        return order, logical_operators
    
    def check_full_clifford(self):
        gates, gates_var_names, gates_as_str = self.full_clifford()
        return self.MAGMA_check_gates(gates, gates_var_names, gates_as_str)
    
    def all_gates(self):
        k = self.k
        H_gates, H_var_names, H_gates_str = self.full_single_qubit_H() 
        S_gates, S_var_names, S_gates_str = self.full_single_qubit_S()
        g_XYZ_gates, g_XYZ_var_names, g_XYZ_str = self.full_single_qubit_gamma_XYZ()
        g_XZY_gates, g_XZY_var_names, g_XZY_str = self.full_single_qubit_gamma_XZY()

        gates_1Q = H_gates + S_gates + g_XYZ_gates + g_XZY_gates
        gates_var_names_1Q = H_var_names[:-1] + ',' + S_var_names[1:-1] + ',' + g_XYZ_var_names[1:-1] + ',' + g_XZY_var_names[1:]
        gates_as_str_1Q = H_gates_str[:-1] + ',' + S_gates_str[1:-1] + ',' + g_XYZ_str[1:-1] + ',' + g_XZY_str[1:]

        if k == 1:
            return gates_1Q, gates_var_names_1Q, gates_as_str_1Q
        
        SWAP_gates, SWAP_var_names, SWAP_gates_str = self.individual_SWAPs()
        CNOT_gates, CNOT_var_names, CNOT_gates_str = self.individual_CNOTs()
        CZ_gates, CZ_var_names, CZ_gates_str = self.individual_CZs()

        gates_2Q = SWAP_gates + CNOT_gates + CZ_gates
        gates_var_names_2Q = SWAP_var_names[:-1] + ',' + CNOT_var_names[1:-1] + ',' + CZ_var_names[1:]
        gates_as_str_2Q = SWAP_gates_str[:-1] + ',' + CNOT_gates_str[1:-1] + ',' + CZ_gates_str[1:]

        gates = gates_1Q + gates_2Q
        gates_var_names = gates_var_names_1Q[:-1] + ',' + gates_var_names_2Q[1:]
        gates_as_str = gates_as_str_1Q[:-1] + ',' + gates_as_str_2Q[1:]

        return gates, gates_var_names, gates_as_str
    
    def check_all_gates_individually(self):
        gates, gates_var_names, gates_as_str = self.all_gates()
        return self.MAGMA_check_gates(gates, gates_var_names, gates_as_str)
    
def codetables_check_all_gates(n,k):
    fileroot = './logical_gates/'
    # filepath = fileroot + f'gates_n{n}k{k}.pkl'
    # with open(filepath, 'rb') as file:
    #     code_auts_data = pickle.load(file)
    filepath = fileroot + f'symp_mats_n{n}k{k}.pkl'
    if os.path.exists(filepath):
        with open(filepath, 'rb') as file:
            symplectic_mats_list = pickle.load(file)
        subgroups = clifford_subgroups(k,symplectic_mats_list)
        order = subgroups.return_order()
        structure = subgroups.return_structure()
        print(f"n{n}k{k}",order)
        print(structure)
        print(subgroups.check_all_gates_individually())
        print('')
    else: 
        print(f"n{n}k{k} - FILE MISSING.")
    return order

def codetables_check_clifford_group(n,k):
    fileroot = './logical_gates/'
    # filepath = fileroot + f'gates_n{n}k{k}.pkl'
    # with open(filepath, 'rb') as file:
    #     code_auts_data = pickle.load(file)
    filepath = fileroot + f'symp_mats_n{n}k{k}.pkl'
    if os.path.exists(filepath):
        with open(filepath, 'rb') as file:
            symplectic_mats_list = pickle.load(file)
        subgroups = clifford_subgroups(k,symplectic_mats_list)
        order = subgroups.return_order()
        structure = subgroups.return_structure()
        print(f"n{n}k{k}",order)
        print(structure)
        print(subgroups.check_full_clifford())
        print('')
    else: 
        print(f"n{n}k{k} - FILE MISSING.")
    return order

def print_code_logicals_data(n,k):
    fileroot = './logical_gates/'
    filepath = fileroot + f'gates_n{n}k{k}.pkl'
    if os.path.exists(filepath):
        with open(filepath, 'rb') as file:
            code_logicals_data = pickle.load(file)
    return code_logicals_data


import numpy as np
import numba as nb

@nb.jit(nb.types.Tuple((nb.int8[:,:],nb.int64[:]))(nb.int8[:,:],nb.int64,nb.int64,nb.int64,nb.int64))
def HowZ2(A,tB,nB,nC,r0):
    pivots = []
    B = A.copy()
    if np.sum(A) == 0:
        return B,np.array(pivots,dtype=nb.int64)
    m = len(B)
    r = r0
    for j in range(nC):
        for t in range(tB):
            ## c is the column of B we are currently looking at
            c = j + t * nB
            iList = [i for i in range(r,m) if B[i,c] > 0]
            if len(iList) > 0:
                i = iList.pop(0)
                pivots.append(c)
                ## found j: if j > r, swap row j with row r
                if i > r:  
                    ## swap using bitflips - more elegant than array indexing
                    B[r] = B[r] ^ B[i]
                    B[i] = B[r] ^ B[i]
                    B[r] = B[r] ^ B[i]
                ## eliminate non-zero entries in column c apart from row r
                for i in [i for i in range(r) if B[i,c] > 0] + iList:     
                    B[i] = B[i] ^ B[r]
                r +=1
    return B,np.array(pivots,dtype=nb.int64)
    
def get_CNOT_circ(A,tB,nB,nC,r0):
    qc = []
    pivots = []
    B = A.copy()
    if np.sum(A) == 0:
        return B,pivots,qc
    m = len(B)
    r = r0
    for j in range(nC):
        for t in range(tB):
            ## c is the column of B we are currently looking at
            c = j + t * nB
            iList = [i for i in range(r,m) if B[i,c] > 0]
            if len(iList) > 0:
                i = iList.pop(0)
                pivots.append(c)
                ## found j: if j > r, swap row j with row r
                if i > r:  
                    ## swap using bitflips - more elegant than array indexing
                    B[r] = B[r] ^ B[i]
                    B[i] = B[r] ^ B[i]
                    B[r] = B[r] ^ B[i]
                    qc.append(('SWAP',(i+1,r+1)))
                ## eliminate non-zero entries in column c apart from row r
                for i in [i for i in range(r) if B[i,c] > 0] + iList:     
                    B[i] = B[i] ^ B[r]
                    qc.append(('CNOT',(i+1,r+1)))
                r +=1
    return B,pivots,qc

# @nb.jit
def blockDims(n,nA=0,tB=1,nC=-1):
    nA = min(n,nA)
    nB = (n - nA) // tB
    if nC < 0 or nC > nB:
        nC = nB 
    return nA,nB,nC

def invRange(n,S):
    '''return list of elements of range(n) NOT in S'''
    return sorted(set(range(n)) - set(S))

def ix_to_perm_mat(ix):
    m = len(ix)
    P = np.zeros((m,m),dtype=int)
    for i in range(m):
        P[ix[i],i] = 1
    return P

def rref_mod2(A,CNOTs=False):
    '''Return Howell matrix form modulo N plus transformation matrix U such that H = U @ A mod N'''
    if A.ndim == 1:
        A = A.reshape(1, -1)
    m,n = A.shape
    A = np.array(A,dtype=np.int8)
    B = np.hstack([A,np.eye(m,dtype=np.int8)])
    nA=0
    tB=1
    nC=-1
    r0=0
    nA,nB,nC = blockDims(n,nA,tB,nC)

    if CNOTs == False:
        HU, pivots = HowZ2(B,tB,nB,nC,r0)
        ix = list(pivots) + invRange(n,pivots)

        H, U = HU[:,:n],HU[:,n:]
        H = H[:,ix]
        P = ix_to_perm_mat(ix)
        
        return H, pivots, U, P
    elif CNOTs: 
        HU, pivots, qc = get_CNOT_circ(B,tB,nB,nC,r0)
        ix = list(pivots) + invRange(n,pivots)
        H, U = HU[:,:n],HU[:,n:]
        H = H[:,ix]

        return qc, H


def rank_mod2(mat):
    """Return rank of binary matrix."""
    _, pivots, _, _ = rref_mod2(mat)
    return len(pivots) 

def is_matrix_full_rank(mat):
    """Checks if matrix is full rank."""
    if mat.ndim == 1:
        return True
    m = mat.shape[0]
    k = rank_mod2(mat)
    return m == k

def inv_mod2(mat):
    _, _, U, P = rref_mod2(mat)
    return (P@U)%2

def diag_mat(n, ix):
    mat = np.zeros((n, n),dtype=int)
    for i in ix:
        mat[i, i] = 1
    return mat

def is_identity_matrix(mat):
    """Checks if matrix is k x k identity."""
    if not mat.shape[0] == mat.shape[1]:
        return False
    return np.all(np.eye(mat.shape[0]) == mat)





from autqec.utils.linalg import *
import pickle
from math import comb
from itertools import combinations

def aut_gens_as_binary_mat(log_circs,k):
    """
    Binary representation: | H | SWAP | CNOT | CZ | S | C(X,X) | Xsqrt |

    Args:
        log_circs (list): list of logical circuits
        k (int): number of logical qubits
    
    Returns:
        binary_vecs_4_auts (np.array): binary representation of logical group generators
    
    """ 
    no_of_gens = len(log_circs)
    kc2 = comb(k,2)
    no_of_cols = 3*k + 4*kc2
    binary_vecs_4_auts = np.zeros((no_of_gens,no_of_cols),dtype=int)
    for g, row in enumerate(binary_vecs_4_auts):
        circ = log_circs[g]
        for op in circ: 
            gate_type, qubits = op
            if gate_type == 'H':
                i = qubits - 1
                col_ind = i
                binary_vecs_4_auts[g, col_ind] = 1
            elif gate_type == 'SWAP':
                i,j = qubits 
                col_ind = comb(k, 2) - comb(k - i + 1, 2) + (j - i - 1) + k 
                binary_vecs_4_auts[g, col_ind] = 1
            elif gate_type == 'CNOT':
                i,j = qubits 
                col_ind = comb(k, 2) - comb(k - i + 1, 2) + (j - i - 1) + k + kc2
                binary_vecs_4_auts[g, col_ind] = 1
            elif gate_type == 'CZ':
                i,j = qubits
                col_ind = comb(k, 2) - comb(k - i + 1, 2) + (j - i - 1) + k + 2*kc2
                binary_vecs_4_auts[g, col_ind] = 1
            elif gate_type == 'S':
                i = qubits - 1
                col_ind = i + k + 3*kc2
                binary_vecs_4_auts[g, col_ind] = 1
            elif gate_type == 'C(X,X)':
                i,j = qubits
                col_ind = comb(k, 2) - comb(k - i + 1, 2) + (j - i - 1) + 2*k + 3*kc2
                binary_vecs_4_auts[g, col_ind] = 1
            elif gate_type == 'Xsqrt':
                i = qubits - 1
                col_ind = i + 2*k + 4*kc2
                binary_vecs_4_auts[g, col_ind] = 1
            else: 
                pass
           
    return binary_vecs_4_auts


def cliff_binary_to_circ(M,k):
    kc2 = comb(k,2)
    no_of_cols = 3*k + 4*kc2
    circs = []
    pairs = list(combinations(range(1, k + 1), 2))
    for g in M: 
        circ = []
        for i in range(no_of_cols): 
            x = g[i]
            if i<k: 
                if x == 1: 
                    circ.append(('H',i+1))
            elif i<k+kc2: 
                if x == 1: 
                    ind = i - k 
                    m,n = pairs[ind]
                    circ.append(('SWAP',(m,n)))
            elif i<k+2*kc2: 
                if x == 1: 
                    ind = i - (k+kc2) 
                    m,n = pairs[ind]
                    circ.append(('CNOT',(m,n)))
            elif i<k+3*kc2:
                if x == 1: 
                    ind = i - (k+2*kc2) 
                    m,n = pairs[ind]
                    circ.append(('CZ',(m,n)))
            elif i<2*k+3*kc2: 
                if x == 1: 
                    ind = i - (k+3*kc2) 
                    circ.append(('S',ind+1))
            elif i<2*k+4*kc2:
                if x == 1: 
                    ind = i - (2*k+3*kc2) 
                    m,n = pairs[ind]
                    circ.append(('C(X,X)',(m,n)))
            elif i>2*k+4*kc2: 
                if x == 1: 
                    print(i,2*k+4*kc2)
                    ind = i - (2*k+4*kc2) 
                    circ.append(('Xsqrt',ind+1))
        circs.append(circ)
    return circs




from collections import Counter
from math import ceil

def permute_by_cycles(ordered_list, permutations):
    permuted_list = ordered_list.copy()
    for cycle in permutations:
        intermediate_list = permuted_list.copy()
        for i, _ in enumerate(cycle):
            old_index = cycle[i-1] - 1
            new_index = cycle[i] - 1 
            permuted_list[new_index] = intermediate_list[old_index]
    return permuted_list

def reduce_triplets_to_qubits(triplets_list):
    triplets = [ceil(x/3) for x in triplets_list]
    # Count occurrences of each element
    elem_counts = Counter(triplets)

    # Assert that each element is repeated three times and is adjacent to identical neighbors
    for element, count in elem_counts.items():
        if count == 3:
            # Find the index of the first occurrence of the element
            first_index = triplets.index(element)
            
            # Check if the next two elements are identical
            if (triplets[first_index + 1] == element and 
                triplets[first_index + 2] == element):
                continue  # Element repeated 3 times and adjacent to identical neighbors
            else:
                print(triplets)
                raise AssertionError(f"Automorphism does not represent a physical operation. Element {element} is not repeated three times consecutively.")
    
    # Reduce the list by getting rid of redundant elements
    reduced_list = list(elem_counts.keys())
    return reduced_list

def reduce_doublets_to_qubits(doublets_list):
    doublets = [ceil(x/2) for x in doublets_list]
    # Count occurrences of each element
    elem_counts = Counter(doublets)

    # Assert that each element is repeated three times and is adjacent to identical neighbors
    for element, count in elem_counts.items():
        if count == 2:
            # Find the index of the first occurrence of the element
            first_index = doublets.index(element)
            
            # Check if the next element is identical
            if (doublets[first_index + 1] == element):
                continue  
            else:
                print(doublets)
                raise AssertionError(f"Automorphism does not represent a physical operation. Element {element} is not repeated two times consecutively.")
    
    # Reduce the list by getting rid of redundant elements
    reduced_list = list(elem_counts.keys())
    return reduced_list

def sort_with_swaps(lst):
    swaps = []
    n = len(lst)
    
    def swap(i, j):
        if i != j:
            lst[i], lst[j] = lst[j], lst[i]
            swaps.append(('SWAP',(i+1,j+1)))
    
    for i in range(n):
        min_idx = i
        for j in range(i+1, n):
            if lst[j] < lst[min_idx]:
                min_idx = j
        
        swap(i, min_idx)
    
    return swaps

def apply_swaps(list_of_lists, instructions):
    def swap(lst, i, j):
        lst[i], lst[j] = lst[j], lst[i]

    for instruction in instructions:
        pair = instruction[1]
        i, j = int(pair[0]) - 1, int(pair[1]) - 1
        swap(list_of_lists, i, j)

    return list_of_lists